"""Initial migration

Revision ID: 9
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.

revision = '9'
down_revision = '8'


def upgrade():
    op.add_column('general', sa.Column('auto_backup_enable', sa.Boolean))
    op.add_column('general', sa.Column('auto_backup_freq', sa.Integer, server_default='24'))
    op.add_column('general', sa.Column('auto_backup_keep_num', sa.Integer, server_default='1'))
    op.add_column('general', sa.Column('auto_backup_dir', sa.Text, server_default=''))


def downgrade():
    pass



"""Initial migration

Revision ID: 7
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.

revision = '7'
down_revision = '6'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    nma = sa.Table('nma', meta, autoload=True)

    for row in conn.execute(nma.select()):
        priority = row.priority

        if isinstance(priority, str):
            try:
                priority = int(priority or 0)
            except ValueError:
                priority = 0

        conn.execute(f'UPDATE nma SET priority = {priority} WHERE nma.id = {row.id}')


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import datetime

import sickrage
from sickrage.core.websocket import WebSocketMessage

MESSAGE = 'notice'
ERROR = 'error'


class Notifications(object):
    """
    A queue of Notification objects.
    """

    def __init__(self):
        self._messages = []
        self._errors = []

    def message(self, title, message=""):
        """
        Add a regular notification to the queue

        title: The title of the notification
        message: The message portion of the notification
        """

        n = Notification(title, message, MESSAGE)
        if not WebSocketMessage('NOTIFICATION', n.data).push():
            self._messages.append(n)

    def error(self, title, message=""):
        """
        Add an error notification to the queue

        title: The title of the notification
        message: The message portion of the notification
        """

        n = Notification(title, message, ERROR)
        if not WebSocketMessage('NOTIFICATION', n.data).push():
            self._errors.append(n)

    def get_notifications(self, remote_ip='127.0.0.1'):
        """
        Return all the available notifications in a list. Marks them all as seen
        as it returns them. Also removes timed out Notifications from the queue.

        Returns: A list of Notification objects
        """

        # filter out expired notifications
        self._errors = [x for x in self._errors if not x.is_expired()]
        self._messages = [x for x in self._messages if not x.is_expired()]

        # return any notifications that haven't been shown to the client already
        return [x.see(remote_ip) for x in self._errors + self._messages if x.is_new(remote_ip)]


class Notification(object):
    """
    Represents a single notification. Tracks its own timeout and a list of which clients have
    seen it before.
    """

    def __init__(self, title, message='', type=None, timeout=None):
        self.title = title
        self.message = str(message) if isinstance(message, Exception) else message
        self._when = datetime.datetime.now()
        self._seen = []
        self._type = type or MESSAGE
        self._timeout = timeout or datetime.timedelta(minutes=1)

    @property
    def data(self):
        return {
            'title': self.title,
            'body': self.message,
            'type': self._type
        }

    def is_new(self, remote_ip='127.0.0.1'):
        """
        Returns True if the notification hasn't been displayed to the current client (aka IP address).
        """
        return remote_ip not in self._seen

    def is_expired(self):
        """
        Returns True if the notification is older than the specified timeout value.
        """
        return datetime.datetime.now() - self._when > self._timeout

    def see(self, remote_ip='127.0.0.1'):
        """
        Returns this notification object and marks it as seen by the client ip
        """
        self._seen.append(remote_ip)
        return self


# class ProgressIndicator:
#     def __init__(self, percentComplete=0, currentStatus=None):
#         if currentStatus is None:
#             currentStatus = {'title': ''}
#         self.percentComplete = percentComplete
#         self.currentStatus = currentStatus


# class ProgressIndicators:
#     _pi = {'massUpdate': [],
#            'massAdd': [],
#            'dailyShowUpdates': []}
#
#     @staticmethod
#     def getIndicator(name):
#         if name not in ProgressIndicators._pi:
#             return []
#
#         # if any of the progress indicators are done take them off the list
#         for curPI in ProgressIndicators._pi[name]:
#             if curPI is not None and curPI.percentComplete() == 100:
#                 ProgressIndicators._pi[name].remove(curPI)
#
#         # return the list of progress indicators associated with this name
#         return ProgressIndicators._pi[name]
#
#     @staticmethod
#     def setIndicator(name, indicator):
#         ProgressIndicators._pi[name].append(indicator)


# class QueueProgressIndicator:
#     """
#     A class used by the UI to show the progress of the queue or a part of it.
#     """
#
#     def __init__(self, name, queueItemList):
#         self.queueItemList = queueItemList
#         self.name = name
#
#     def numTotal(self):
#         return len(self.queueItemList)
#
#     def numFinished(self):
#         return len([x for x in self.queueItemList if not x.is_in_queue()])
#
#     def numRemaining(self):
#         return len([x for x in self.queueItemList if x.is_in_queue()])
#
#     def nextName(self):
#         for cur_item in self.queue_items:
#             if cur_item in self.queueItemList:
#                 return cur_item.name
#
#         return "Unknown"
#
#     def percentComplete(self):
#         numFinished = self.numFinished()
#         numTotal = self.numTotal()
#
#         if numTotal == 0:
#             return 0
#         else:
#             return int(float(numFinished) / float(numTotal) * 100)


class LoadingTVShow:
    def __init__(self, dir):
        self.dir = dir
        self.show = None



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


from operator import itemgetter
from urllib.parse import quote

import sickrage
from sickrage.core.enums import SeriesProviderID
from sickrage.series_providers import SeriesProvider


class SeriesProviderActors(list):
    """Holds all Actor instances for a show
    """
    pass


class SeriesProviderActor(dict):
    """Represents a single actor. Should contain..

    id,
    image,
    name,
    role,
    sortorder
    """

    def __repr__(self):
        return "<Actor \"{}\">".format(self.get("name"))


class TheTVDB(SeriesProvider):
    """Create easy-to-use interface to name of season/episode name
    """

    def __init__(self):
        super(TheTVDB, self).__init__(SeriesProviderID.THETVDB)
        self.trakt_id = 'tvdb'
        self.xem_origin = 'tvdb'
        self.icon = 'thetvdb16.png'
        self.show_url = 'http://thetvdb.com/?tab=series&id='
        self.dvd_order = False

    def search(self, query, language='eng'):
        """
        This searches TheTVDB.com for the series by name and returns the result list
        """

        sickrage.app.log.debug(f"Searching for show using query term: {query}")

        search_results = sickrage.app.api.series_provider.search(provider=self.slug, query=quote(query), language=language)
        if not search_results or 'error' in search_results:
            sickrage.app.log.debug(f'Series search using query term {query} returned zero results, cannot find series on {self.name}')

        return search_results

    def search_by_id(self, remote_id, language='eng'):
        """
        This searches TheTVDB.com for the seriesid, imdbid, or zap2itid  and returns the result list
        """

        sickrage.app.log.debug(f"Searching for show using remote id: {remote_id}")

        if not isinstance(remote_id, int):
            remote_id = quote(remote_id)

        search_result = sickrage.app.api.series_provider.search_by_id(provider=self.slug, remote_id=remote_id, language=language)
        if not search_result or 'error' in search_result:
            sickrage.app.log.debug(f'Series search using remote id {remote_id} returned zero results, cannot find series on {self.name}')

        return search_result

    def get_series_info(self, sid, language='eng', dvd_order=False, enable_cache=True):
        """
        Takes a series id, gets the episodes URL and parses the TVDB
        """

        # check if series is in cache
        if sid in self.cache and enable_cache:
            search_result = self.cache[sid]
            if search_result:
                return search_result

        # get series data
        sickrage.app.log.debug(f"[{sid}]: Getting series info from {self.name}")

        resp = sickrage.app.api.series_provider.get_series_info(provider=self.slug, series_id=sid, language=language)
        if not resp or 'error' in resp:
            sickrage.app.log.debug(f"[{sid}]: Unable to get series info from {self.name}")
            return None

        # add season data to cache
        for season in resp['seasons']:
            season_number = int(float(season.get('seasonNumber')))

            for k, v in season.items():
                self.cache.add_season_data(sid, season_number, k, v)

        # add series data to cache
        [self.cache.add_show_data(sid, k, v) for k, v in resp.items() if k != 'seasons']

        # get season and episode data
        sickrage.app.log.debug(f'[{sid}]: Getting episode data from {self.name}')

        season_type = 'dvd' if dvd_order else 'official'
        resp = sickrage.app.api.series_provider.get_episodes_info(provider=self.slug, series_id=sid, season_type=season_type, language=language)
        if not resp or 'error' in resp:
            sickrage.app.log.debug(f"[{sid}]: Unable to get episode data from {self.name}")
            return None

        # add episode data to cache
        episode_incomplete = False
        for episode in resp:
            season_number, episode_number = episode.get('seasonNumber'), episode.get('episodeNumber')
            if season_number is None or episode_number is None:
                episode_incomplete = True
                continue

            season_number = int(float(season_number))
            episode_number = int(float(episode_number))

            for k, v in episode.items():
                self.cache.add_episode_data(sid, season_number, episode_number, k, v)

        if episode_incomplete:
            sickrage.app.log.debug(f"{sid}: Series has incomplete season/episode numbers")

        # set last updated
        # self.cache.add_show_data(sid, 'last_updated', int(time.mktime(datetime.now().timetuple())))

        return self.cache[int(sid)]

    def image_types(self):
        return {
            'series': {
                'banner': 1,
                'poster': 2,
                'fanart': 3
            },
            'season': {
                'banner': 6,
                'poster': 7,
                'fanart': 8
            }
        }

    def images(self, sid, key_type='poster', season=None, language='eng'):
        sickrage.app.log.debug(f'Getting {key_type} images for {sid}')

        images = []

        series_info = self.get_series_info(sid=sid, language=language)
        if not series_info:
            return []

        season_map = {}
        for season_number in series_info:
            season_map[season_number] = series_info[season_number]['id']

        for item in sorted(series_info.artworks, key=itemgetter("score"), reverse=True):
            if season and season_map[season] == item['seasonId'] and item['type'] == self.image_types()['season'][key_type]:
                images.append(item)
            elif not season and item['type'] == self.image_types()['series'][key_type]:
                images.append(item)

        if not images and key_type == 'poster':
            if season:
                image_url = series_info[season].imageUrl
                if image_url != '':
                    images.append({'image': image_url})
            else:
                image_url = series_info.imageUrl
                if image_url != '':
                    images.append({'image': image_url})

        return images

    def updates(self, since):
        resp = sickrage.app.api.series_provider.updates(provider=self.slug, since=since)
        if resp and 'error' not in resp:
            return resp

    def languages(self):
        resp = sickrage.app.api.series_provider.languages(provider=self.slug)
        if not resp or 'error' in resp:
            return {}

        return sorted(resp, key=lambda i: i['name'])

    def __repr__(self):
        return repr(self.cache)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

__version__ = "10.0.71.dev2"
__install_type__ = "pip"

import sys

# sickrage requires python 3.6+
if sys.version_info < (3, 6, 0):
    sys.exit("Sorry, SiCKRAGE requires Python 3.6+")

import argparse
import atexit
import gettext
import multiprocessing
import os
import pathlib
import re
import site
import subprocess
import threading
import time
import traceback
import pkg_resources

# pywin32 for windows service
try:
    import win32api
    import win32serviceutil
    import win32evtlogutil
    import win32event
    import win32service
    import win32ts
    import servicemanager
    from win32com.shell import shell, shellcon
except ImportError:
    if __install_type__ == 'windows':
        sys.exit("Sorry, SiCKRAGE requires Python module PyWin32.")

from signal import SIGTERM

app = None

MAIN_DIR = os.path.abspath(os.path.realpath(os.path.expanduser(os.path.dirname(os.path.dirname(__file__)))))
PROG_DIR = os.path.abspath(os.path.realpath(os.path.expanduser(os.path.dirname(__file__))))
LOCALE_DIR = os.path.join(PROG_DIR, 'locale')
CHANGELOG_FILE = os.path.join(MAIN_DIR, 'CHANGELOG.md')
REQS_FILE = os.path.join(MAIN_DIR, 'requirements.txt')
CHECKSUM_FILE = os.path.join(PROG_DIR, 'checksums.md5')
AUTO_PROCESS_TV_CFG_FILE = os.path.join(*[PROG_DIR, 'autoProcessTV', 'autoProcessTV.cfg'])

# add sickrage libs path to python system path
LIBS_DIR = os.path.join(PROG_DIR, 'libs')
if not (LIBS_DIR in sys.path) and not getattr(sys, 'frozen', False):
    sys.path, remainder = sys.path[:1], sys.path[1:]
    site.addsitedir(LIBS_DIR)
    sys.path.extend(remainder)

# set system default language
gettext.install('messages', LOCALE_DIR, codeset='UTF-8', names=["ngettext"])

if __install_type__ == 'windows':
    class SiCKRAGEService(win32serviceutil.ServiceFramework):
        _svc_name_ = "SiCKRAGE"
        _svc_display_name_ = "SiCKRAGE"
        _svc_description_ = (
            "Automated video library manager for TV shows. "
            'Set to "automatic" to start the service at system startup. '
            "You may need to login with a real user account when you need "
            "access to network shares."
        )

        if hasattr(sys, "frozen"):
            _exe_name_ = "SiCKRAGE.exe"

        def __init__(self, args):
            win32serviceutil.ServiceFramework.__init__(self, args)
            self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

        def SvcDoRun(self):
            msg = "SiCKRAGE-service %s" % __version__
            self.Logger(servicemanager.PYS_SERVICE_STARTED, msg + " has started")
            start()
            self.Logger(servicemanager.PYS_SERVICE_STOPPED, msg + " has stopped")

        def SvcStop(self):
            if app:
                app.shutdown()

            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
            win32event.SetEvent(self.hWaitStop)

        def Logger(self, state, msg):
            win32evtlogutil.ReportEvent(
                self._svc_display_name_, state, 0, servicemanager.EVENTLOG_INFORMATION_TYPE, (self._svc_name_, msg)
            )


class Daemon(object):
    """
    Usage: subclass the Daemon class
    """

    def __init__(self, pidfile, working_dir="/"):
        self.stdin = getattr(os, 'devnull', '/dev/null')
        self.stdout = getattr(os, 'devnull', '/dev/null')
        self.stderr = getattr(os, 'devnull', '/dev/null')
        self.pidfile = pidfile
        self.working_dir = working_dir
        self.pid = None

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                os._exit(0)
        except OSError as e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                os._exit(0)
        except OSError as e:
            sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdin = sys.__stdin__
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(self.stdin, 'r')
        so = open(self.stdout, 'a+')
        se = open(self.stderr, 'a+')
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        self.pid = os.getpid()
        open(self.pidfile, 'w+').write("%s\n" % self.pid)

    def delpid(self):
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)

    def start(self):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if pid:
            message = "pidfile %s already exist. Daemon already running?\n"
            sys.stderr.write(message % self.pidfile)
            sys.exit(1)

        # Start the daemon
        self.daemonize()

    def stop(self):
        """
        Stop the daemon
        """

        # Get the pid from the pidfile
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            err = str(err)
            if err.find("No such process") > 0:
                self.delpid()
            else:
                sys.exit(1)


def version():
    # Return the version number
    return __version__


def install_type():
    # Return the install type
    if not __install_type__ and os.path.isdir(os.path.join(MAIN_DIR, '.git')):
        return 'git'
    else:
        return __install_type__ or 'source'


def changelog():
    # Return contents of CHANGELOG.md
    with open(CHANGELOG_FILE) as f:
        return f.read()


def check_requirements():
    if os.path.exists(REQS_FILE):
        with open(REQS_FILE) as f:
            for line in f.readlines():
                try:
                    req_name, req_version = line.strip().split('==', 1)
                    if 'python_version' in req_version:
                        m = re.search('(\d+.\d+.\d+).*(\d+.\d+)', req_version)
                        req_version = m.group(1)
                        python_version = m.group(2)
                        python_version_major = int(python_version.split('.')[0])
                        python_version_minor = int(python_version.split('.')[1])
                        if sys.version_info.major == python_version_major and sys.version_info.minor != python_version_minor:
                            continue

                    if not pkg_resources.get_distribution(req_name).version == req_version:
                        print('Updating requirement {} to {}'.format(req_name, req_version))
                        subprocess.check_call([sys.executable, "-m", "pip", "install", "--no-deps", "--no-cache-dir", line.strip()])
                except pkg_resources.DistributionNotFound:
                    print('Installing requirement {}'.format(line.strip()))
                    subprocess.check_call([sys.executable, "-m", "pip", "install", "--no-deps", "--no-cache-dir", line.strip()])
                except ValueError:
                    continue


def verify_checksums(remove_unverified=False):
    valid_files = []
    exempt_files = [pathlib.Path(__file__), pathlib.Path(CHECKSUM_FILE), pathlib.Path(AUTO_PROCESS_TV_CFG_FILE)]

    if not os.path.exists(CHECKSUM_FILE):
        return

    with open(CHECKSUM_FILE, "rb") as fp:
        for line in fp.readlines():
            file, checksum = line.decode().strip().split(' = ')
            full_filename = pathlib.Path(MAIN_DIR).joinpath(file)
            valid_files.append(full_filename)

    for root, dirs, files in os.walk(PROG_DIR):
        for file in files:
            full_filename = pathlib.Path(root).joinpath(file)

            if full_filename in exempt_files or full_filename.suffix == '.pyc':
                continue

            if full_filename not in valid_files and PROG_DIR in str(full_filename):
                try:
                    if remove_unverified:
                        print('Found unverified file {}, removed!'.format(full_filename))
                        full_filename.unlink()
                    else:
                        print('Found unverified file {}, you should delete this file manually!'.format(full_filename))
                except OSError:
                    print('Unable to delete unverified filename {} during checksum verification, you should delete this file manually!'.format(full_filename))


def handle_windows_service():
    if hasattr(sys, "frozen") and win32ts.ProcessIdToSessionId(win32api.GetCurrentProcessId()) == 0:
        servicemanager.Initialize()
        servicemanager.PrepareToHostSingle(SiCKRAGEService)
        servicemanager.StartServiceCtrlDispatcher()
        return True

    if len(sys.argv) > 1 and sys.argv[1] in ("install", "update", "remove", "start", "stop", "restart", "debug"):
        win32serviceutil.HandleCommandLine(SiCKRAGEService)
        del sys.argv[1]
        return True


def main():
    multiprocessing.freeze_support()

    # set thread name
    threading.current_thread().name = 'MAIN'

    # fix threading time bug
    time.strptime("2012", "%Y")

    if __install_type__ == 'windows':
        if not handle_windows_service():
            start()
    else:
        start()


def start():
    global app

    parser = argparse.ArgumentParser(
        prog='sickrage',
        description='Automated video library manager for TV shows'
    )

    parser.add_argument('-v', '--version',
                        action='version',
                        version=version())
    parser.add_argument('-d', '--daemon',
                        action='store_true',
                        help='Run as a daemon (*NIX ONLY)')
    parser.add_argument('-q', '--quiet',
                        action='store_true',
                        help='Disables logging to CONSOLE')
    parser.add_argument('-p', '--port',
                        default=0,
                        type=int,
                        help='Override default/configured port to listen on')
    parser.add_argument('-H', '--host',
                        default='',
                        help='Override default/configured host to listen on')
    parser.add_argument('--dev',
                        action='store_true',
                        help='Enable developer mode')
    parser.add_argument('--debug',
                        action='store_true',
                        help='Enable debugging')
    parser.add_argument('--datadir',
                        default=os.path.abspath(os.path.join(os.path.expanduser("~"), '.sickrage')),
                        help='Overrides data folder for database, config, cache and logs (specify full path)')
    parser.add_argument('--config',
                        default='config.ini',
                        help='Overrides config filename (specify full path and filename if outside datadir path)')
    parser.add_argument('--pidfile',
                        default='sickrage.pid',
                        help='Creates a PID file (specify full path and filename if outside datadir path)')
    parser.add_argument('--no_clean',
                        action='store_true',
                        help='Suppress cleanup of files not present in checksum.md5')
    parser.add_argument('--nolaunch',
                        action='store_true',
                        help='Suppress launching web browser on startup')
    parser.add_argument('--disable_updates',
                        action='store_true',
                        help='Disable application updates')
    parser.add_argument('--web_root',
                        default='',
                        type=str,
                        help='Overrides URL web root')
    parser.add_argument('--db_type',
                        default='sqlite',
                        help='Database type: sqlite or mysql')
    parser.add_argument('--db_prefix',
                        default='sickrage',
                        help='Database prefix you want prepended to database table names')
    parser.add_argument('--db_host',
                        default='localhost',
                        help='Database hostname (not used for sqlite)')
    parser.add_argument('--db_port',
                        default='3306',
                        help='Database port number (not used for sqlite)')
    parser.add_argument('--db_username',
                        default='sickrage',
                        help='Database username (not used for sqlite)')
    parser.add_argument('--db_password',
                        default='sickrage',
                        help='Database password (not used for sqlite)')

    # Parse startup args
    args = parser.parse_args()

    # check requirements
    # if install_type() not in ['windows', 'synology', 'docker', 'qnap', 'readynas', 'pip']:
    #     check_requirements()

    # verify file checksums, remove unverified files
    # verify_checksums(remove_unverified=not args.no_clean)

    try:
        from sickrage.core import Core
        app = Core()
    except ImportError:
        sys.exit("Sorry, SiCKRAGE requirements need to be installed.")

    try:
        app.quiet = args.quiet
        app.web_host = args.host
        app.web_port = int(args.port)
        app.web_root = args.web_root.lstrip('/').rstrip('/')
        app.no_launch = args.nolaunch
        app.disable_updates = args.disable_updates
        app.developer = args.dev
        app.db_type = args.db_type
        app.db_prefix = args.db_prefix
        app.db_host = args.db_host
        app.db_port = args.db_port
        app.db_username = args.db_username
        app.db_password = args.db_password
        app.debug = args.debug
        app.data_dir = os.path.abspath(os.path.realpath(os.path.expanduser(args.datadir)))
        app.cache_dir = os.path.abspath(os.path.realpath(os.path.join(app.data_dir, 'cache')))
        app.config_file = args.config
        daemonize = (False, args.daemon)[not sys.platform == 'win32']
        pid_file = args.pidfile

        if not os.path.isabs(app.config_file):
            app.config_file = os.path.join(app.data_dir, app.config_file)

        if not os.path.isabs(pid_file):
            pid_file = os.path.join(app.data_dir, pid_file)

        # add sickrage module to python system path
        if not (PROG_DIR in sys.path) and not getattr(sys, 'frozen', False):
            sys.path, remainder = sys.path[:1], sys.path[1:]
            site.addsitedir(PROG_DIR)
            sys.path.extend(remainder)

        # Make sure that we can create the data dir
        if not os.access(app.data_dir, os.F_OK):
            try:
                os.makedirs(app.data_dir, 0o744)
            except os.error:
                sys.exit("Unable to create data directory '" + app.data_dir + "'")

        # Make sure we can write to the data dir
        if not os.access(app.data_dir, os.W_OK):
            sys.exit("Data directory must be writeable '" + app.data_dir + "'")

        # Make sure that we can create the cache dir
        if not os.access(app.cache_dir, os.F_OK):
            try:
                os.makedirs(app.cache_dir, 0o744)
            except os.error:
                sys.exit("Unable to create cache directory '" + app.cache_dir + "'")

        # Make sure we can write to the cache dir
        if not os.access(app.cache_dir, os.W_OK):
            sys.exit("Cache directory must be writeable '" + app.cache_dir + "'")

        # daemonize if requested
        if daemonize:
            app.no_launch = True
            app.quiet = True
            app.daemon = Daemon(pid_file, app.data_dir)
            app.daemon.daemonize()
            app.pid = app.daemon.pid

        app.start()

        from tornado.ioloop import IOLoop
        IOLoop.current().start()
    except (SystemExit, KeyboardInterrupt):
        if app:
            app.shutdown()
    except Exception as e:
        try:
            # attempt to send exception to sentry
            import sentry_sdk
            sentry_sdk.capture_exception(e)
        except ImportError:
            pass

        traceback.print_exc()


if __name__ == '__main__':
    main()



"""Initial migration

Revision ID: 23
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""

import babelfish
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '23'
down_revision = '22'


def upgrade():
    pass


def downgrade():
    pass



import datetime
import os

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509 import ExtensionNotFound
from cryptography.x509.oid import NameOID

import sickrage


def create_https_certificates(ssl_cert, ssl_key):
    """This function takes a domain name as a parameter and then creates a certificate and key with the
    domain name(replacing dots by underscores), finally signing the certificate using specified CA and
    returns the path of key and cert files. If you are yet to generate a CA then check the top comments"""

    # Generate our key
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )

    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, 'SiCKRAGE')
    ])

    # path_len=0 means this cert can only sign itself, not other certs.
    basic_contraints = x509.BasicConstraints(ca=True, path_length=0)
    now = datetime.datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(key.public_key())
            .serial_number(1000)
            .not_valid_before(now)
            .not_valid_after(now + datetime.timedelta(days=10 * 365))
            .add_extension(basic_contraints, False)
            # .add_extension(san, False)
            .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    with open(ssl_key, 'wb') as key_out:
        key_out.write(key_pem)

    with open(ssl_cert, 'wb') as cert_out:
        cert_out.write(cert_pem)

    return True


def is_certificate_valid(cert_file):
    if not os.path.exists(cert_file):
        return

    with open(cert_file, 'rb') as f:
        cert_pem = f.read()

    cert = x509.load_pem_x509_certificate(cert_pem, default_backend())
    issuer = cert.issuer.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0]
    subject = cert.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0]

    if 'ZeroSSL' not in issuer.value:
        return False

    if subject.value != f'*.{sickrage.app.config.general.server_id}.sickrage.direct':
        return False

    try:
        ext = cert.extensions.get_extension_for_class(x509.SubjectAlternativeName).value
        sans = ext.get_values_for_type(x509.DNSName)

        domains = [f'*.{sickrage.app.config.general.server_id}.sickrage.direct']

        for domain in sans:
            if domain not in domains:
                return False
    except ExtensionNotFound:
        return False

    return True


def certificate_needs_renewal(cert_file):
    if not os.path.exists(cert_file):
        return

    with open(cert_file, 'rb') as f:
        cert_pem = f.read()

    cert_info = x509.load_pem_x509_certificate(cert_pem, default_backend())
    expiry_date = cert_info.not_valid_after
    time_left = expiry_date.date() - datetime.date.today()

    return time_left.days < 1



"""Initial migration

Revision ID: 8
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.

revision = '8'
down_revision = '7'


def upgrade():
    op.add_column('general', sa.Column('update_video_metadata', sa.Boolean, server_default='true'))


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

from google.protobuf.json_format import MessageToDict
from pika.exceptions import ChannelWrongStateError, StreamLostError
from tornado.ioloop import IOLoop

import sickrage
from sickrage.core.amqp import AMQPBase
from sickrage.core.amqp.protos.announcement_v1_pb2 import CreatedAnnouncementResponse, DeletedAnnouncementResponse
from sickrage.core.amqp.protos.network_timezone_v1_pb2 import SavedNetworkTimezoneResponse, DeletedNetworkTimezoneResponse
from sickrage.core.amqp.protos.search_provider_url_v1_pb2 import SavedSearchProviderUrlResponse
from sickrage.core.amqp.protos.server_certificate_v1_pb2 import SavedServerCertificateResponse
from sickrage.core.amqp.protos.updates_v1_pb2 import UpdatedAppResponse


class AMQPConsumer(AMQPBase):
    def __init__(self):
        super(AMQPConsumer, self).__init__()

    @property
    def events(self):
        return {
            'server_ssl_certificate.saved': {
                'event_msg': SavedServerCertificateResponse(),
                'event_cmd': sickrage.app.wserver.load_ssl_certificate,
            },
            'network_timezone.saved': {
                'event_msg': SavedNetworkTimezoneResponse(),
                'event_cmd': sickrage.app.tz_updater.update_network_timezone,
            },
            'network_timezone.deleted': {
                'event_msg': DeletedNetworkTimezoneResponse(),
                'event_cmd': sickrage.app.tz_updater.delete_network_timezone,
            },
            'search_provider_url.saved': {
                'event_msg': SavedSearchProviderUrlResponse(),
                'event_cmd': sickrage.app.search_providers.update_url,
            },
            'app.updated': {
                'event_msg': UpdatedAppResponse(),
                'event_cmd': sickrage.app.version_updater.task,
            },
            'announcement.created': {
                'event_msg': CreatedAnnouncementResponse(),
                'event_cmd': sickrage.app.announcements.add,
            },
            'announcement.deleted': {
                'event_msg': DeletedAnnouncementResponse(),
                'event_cmd': sickrage.app.announcements.clear,
            },
        }

    def on_channel_open(self, channel):
        self._channel = channel
        self._channel.basic_qos(callback=self.on_qos_applied, prefetch_count=self._prefetch_count)

    def on_qos_applied(self, method):
        self.start_consuming()

    def on_message(self, unused_channel, basic_deliver, properties, body):
        try:
            if basic_deliver.exchange in self.events:
                event = self.events[basic_deliver.exchange]

                message = event['event_msg']
                message.ParseFromString(body)
                message_kwargs = MessageToDict(message, including_default_value_fields=True, preserving_proto_field_name=True)

                sickrage.app.log.debug(
                    f"Received AMQP event: {basic_deliver.exchange} :: {message_kwargs!r}"
                )

                IOLoop.current().spawn_callback(event['event_cmd'], **message_kwargs)
        except Exception as e:
            sickrage.app.log.debug(f"AMQP exchange: {basic_deliver.exchange} message caused an exception: {e!r}")
        finally:
            self._channel.basic_ack(basic_deliver.delivery_tag)

    def start_consuming(self):
        sickrage.app.log.info('Connected to SiCKRAGE AMQP server')

        try:
            self._consumer_tag = self._channel.basic_consume(
                on_message_callback=self.on_message,
                queue=f'{sickrage.app.config.user.sub_id}.{sickrage.app.config.general.server_id}',
            )
        except (ChannelWrongStateError, StreamLostError):
            sickrage.app.log.debug('AMQP channel error, attempting to reconnect')
            IOLoop.current().call_later(5, self.reconnect)


"""Initial migration

Revision ID: 22
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""

import babelfish
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '22'
down_revision = '21'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    tv_shows = sa.Table('tv_shows', meta, autoload=True)

    with op.get_context().begin_transaction():
        for row in conn.execute(tv_shows.select()):
            if len(row.lang) == 2:
                lang = babelfish.Language.fromalpha2(row.lang)
                conn.execute(f'UPDATE tv_shows SET lang = "{lang.alpha3}" WHERE tv_shows.series_id = {row.series_id}')


def downgrade():
    pass



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import datetime
import os
from collections import OrderedDict
from time import sleep
from urllib.parse import unquote_plus, quote_plus

from apscheduler.triggers.date import DateTrigger
from tornado.escape import json_encode
from tornado.httputil import url_concat
from tornado.web import authenticated

import sickrage
from sickrage.clients import get_client_instance
from sickrage.clients.nzb.sabnzbd import SabNZBd
from sickrage.core.common import Overview, Quality, Qualities
from sickrage.core.enums import SeriesProviderID, TorrentMethod, NzbMethod
from sickrage.core.exceptions import (
    AnidbAdbaConnectionException,
    CantRefreshShowException,
    CantUpdateShowException,
    CantRemoveShowException,
    EpisodeDeletedException,
    EpisodeNotFoundException,
    MultipleEpisodesInDatabaseException
)
from sickrage.core.helpers import clean_url, clean_host, clean_hosts, get_disk_space_usage
from sickrage.core.helpers.anidb import get_release_groups_for_anime
from sickrage.core.helpers.srdatetime import SRDateTime
from sickrage.core.queues import TaskStatus
from sickrage.core.queues.search import FailedSearchTask, ManualSearchTask
from sickrage.core.scene_numbering import (
    get_scene_numbering_for_show,
    get_xem_numbering_for_show,
    get_scene_absolute_numbering_for_show,
    get_xem_absolute_numbering_for_show,
    set_scene_numbering,
    get_scene_absolute_numbering,
    get_scene_numbering
)
from sickrage.core.traktapi import TraktAPI
from sickrage.core.tv.show.helpers import find_show, get_show_list
from sickrage.core.webserver.handlers.base import BaseHandler
from sickrage.subtitles import Subtitles


class HomeHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show_list = [x for x in get_show_list() if not sickrage.app.show_queue.is_being_removed(x.series_id)]

        if not len(show_list):
            return self.redirect('/home/addShows/')

        show_lists = OrderedDict({
            'Shows': [x for x in show_list if x.anime is False],
            'Anime': [x for x in show_list if x.anime is True]
        })

        return self.render('home/index.mako',
                           title="Home",
                           header="Show List",
                           topmenu="home",
                           showlists=show_lists,
                           controller='home',
                           action='index')

    def statistics(self):
        show_stat = {}

        overall_stats = {
            'episodes': {
                'downloaded': 0,
                'snatched': 0,
                'total': 0,
            },
            'shows': {
                'active': len([show for show in get_show_list() if show.paused == 0 and show.status.lower() == 'continuing']),
                'total': len(get_show_list()),
            },
            'total_size': 0
        }

        for show in get_show_list():
            if sickrage.app.show_queue.is_being_added(show.series_id) or sickrage.app.show_queue.is_being_removed(show.series_id):
                show_stat[show.series_id] = {
                    'ep_airs_next': datetime.date.min,
                    'ep_airs_prev': datetime.date.min,
                    'ep_snatched': 0,
                    'ep_downloaded': 0,
                    'ep_total': 0,
                    'total_size': 0
                }
            else:
                show_stat[show.series_id] = {
                    'ep_airs_next': show.airs_next or datetime.date.min,
                    'ep_airs_prev': show.airs_prev or datetime.date.min,
                    'ep_snatched': show.episodes_snatched or 0,
                    'ep_downloaded': show.episodes_downloaded or 0,
                    'ep_total': len(show.episodes),
                    'total_size': show.total_size or 0
                }

            overall_stats['episodes']['snatched'] += show_stat[show.series_id]['ep_snatched']
            overall_stats['episodes']['downloaded'] += show_stat[show.series_id]['ep_downloaded']
            overall_stats['episodes']['total'] += show_stat[show.series_id]['ep_total']
            overall_stats['total_size'] += show_stat[show.series_id]['total_size']

        return show_stat, overall_stats


class ShowProgressHandler(BaseHandler):
    def get(self, *args, **kwargs):
        series_id = self.get_argument('show-id')

        show = find_show(int(series_id))
        if not show:
            return

        episodes_snatched = show.episodes_snatched
        episodes_downloaded = show.episodes_downloaded
        episodes_total = show.episodes_total
        progressbar_percent = int(episodes_downloaded * 100 / episodes_total if episodes_total > 0 else 1)

        progress_text = '?'
        progress_tip = _("no data")
        if episodes_total != 0:
            progress_text = str(episodes_downloaded)
            progress_tip = _("Downloaded: ") + str(episodes_downloaded)
            if episodes_snatched > 0:
                progress_text = progress_text + "+" + str(episodes_snatched)
                progress_tip = progress_tip + "&#013;" + _("Snatched: ") + str(episodes_snatched)

            progress_text = progress_text + " / " + str(episodes_total)
            progress_tip = progress_tip + "&#013;" + _("Total: ") + str(episodes_total)

        return json_encode({'progress_text': progress_text, 'progress_tip': progress_tip, 'progressbar_percent': progressbar_percent})


class IsAliveHandler(BaseHandler):
    def get(self, *args, **kwargs):
        self.set_header('Content-Type', 'text/javascript')

        srcallback = self.get_argument('srcallback')

        if not srcallback:
            return _("Error: Unsupported Request. Send jsonp request with 'srcallback' variable in the query string.")

        return '{}({})'.format(srcallback, {"msg": str(sickrage.app.pid) if sickrage.app.started else 'nope'})


class TestSABnzbdHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_url(self.get_argument('host'))
        username = self.get_argument('username')
        password = self.get_argument('password')
        apikey = self.get_argument('apikey')

        connection, acces_msg = SabNZBd.get_sab_access_method(host)

        if connection:
            authed, auth_msg = SabNZBd.test_authentication(host, username, password, apikey)
            if authed:
                return _('Success. Connected and authenticated')
            return _('Authentication failed. SABnzbd expects {access!r} as authentication method, {auth}'.format(access=acces_msg, auth=auth_msg))
        return _('Unable to connect to host')


class TestSynologyDSMHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_url(self.get_argument('host'))
        nzb_method = self.get_argument('nzb_method')
        username = self.get_argument('username')
        password = self.get_argument('password')

        client = get_client_instance(NzbMethod[nzb_method].value, client_type='nzb')
        __, access_msg = client(host, username, password).test_authentication()
        return access_msg


class TestTorrentHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_url(self.get_argument('host'))
        torrent_method = self.get_argument('torrent_method')
        username = self.get_argument('username')
        password = self.get_argument('password')

        client = get_client_instance(TorrentMethod[torrent_method].value, client_type='torrent')
        __, access_msg = client(host, username, password).test_authentication()
        return access_msg


class TestFreeMobileHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        freemobile_id = self.get_argument('freemobile_id')
        freemobile_apikey = self.get_argument('freemobile_apikey')

        result, message = sickrage.app.notification_providers['freemobile'].test_notify(freemobile_id, freemobile_apikey)
        if result:
            return _('SMS sent successfully')
        return _('Problem sending SMS: ') + message


class TestTelegramHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        telegram_id = self.get_argument('telegram_id')
        telegram_apikey = self.get_argument('telegram_apikey')

        result, message = sickrage.app.notification_providers['telegram'].test_notify(telegram_id, telegram_apikey)
        if result:
            return _('Telegram notification succeeded. Check your Telegram clients to make sure it worked')
        return _('Error sending Telegram notification: {message}').format(message=message)


class TestJoinHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        join_id = self.get_argument('join_id')
        join_apikey = self.get_argument('join_apikey')

        result, message = sickrage.app.notification_providers['join'].test_notify(join_id, join_apikey)
        if result:
            return _('Join notification succeeded. Check your Join clients to make sure it worked')
        return _('Error sending Join notification: {message}').format(message=message)


class TestGrowlHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'), default_port=23053)
        password = self.get_argument('password')

        result = sickrage.app.notification_providers['growl'].test_notify(host, password)
        if password is None or password == '':
            pw_append = ''
        else:
            pw_append = _(' with password: ') + password

        if result:
            return _('Registered and tested Growl successfully ') + unquote_plus(host) + pw_append
        return _('Registration and testing of Growl failed ') + unquote_plus(host) + pw_append


class TestProwlHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        prowl_apikey = self.get_argument('prowl_apikey')
        prowl_priority = self.get_argument('prowl_priority')

        result = sickrage.app.notification_providers['prowl'].test_notify(prowl_apikey, prowl_priority)
        if result:
            return _('Test prowl notice sent successfully')
        return _('Test prowl notice failed')


class TestBoxcar2Handler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        accesstoken = self.get_argument('accesstoken')

        result = sickrage.app.notification_providers['boxcar2'].test_notify(accesstoken)
        if result:
            return _('Boxcar2 notification succeeded. Check your Boxcar2 clients to make sure it worked')
        return _('Error sending Boxcar2 notification')


class TestPushoverHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        user_key = self.get_argument('userKey')
        api_key = self.get_argument('apiKey')

        result = sickrage.app.notification_providers['pushover'].test_notify(user_key, api_key)
        if result:
            return _('Pushover notification succeeded. Check your Pushover clients to make sure it worked')
        return _('Error sending Pushover notification')


class TwitterStep1Handler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return sickrage.app.notification_providers['twitter']._get_authorization()


class TwitterStep2Handler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        key = self.get_argument('key')

        result = sickrage.app.notification_providers['twitter']._get_credentials(key)
        sickrage.app.log.info("result: " + str(result))
        if result:
            return _('Key verification successful')
        return _('Unable to verify key')


class TestTwitterHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        result = sickrage.app.notification_providers['twitter'].test_notify()
        if result:
            return _('Tweet successful, check your twitter to make sure it worked')
        return _('Error sending tweet')


class TestTwilioHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        account_sid = self.get_argument('account_sid')
        auth_token = self.get_argument('auth_token')
        phone_sid = self.get_argument('phone_sid')
        to_number = self.get_argument('to_number')

        if not sickrage.app.notification_providers['twilio'].account_regex.match(account_sid):
            return _('Please enter a valid account sid')

        if not sickrage.app.notification_providers['twilio'].auth_regex.match(auth_token):
            return _('Please enter a valid auth token')

        if not sickrage.app.notification_providers['twilio'].phone_regex.match(phone_sid):
            return _('Please enter a valid phone sid')

        if not sickrage.app.notification_providers['twilio'].number_regex.match(to_number):
            return _('Please format the phone number as "+1-###-###-####"')

        result = sickrage.app.notification_providers['twilio'].test_notify()
        if result:
            return _('Authorization successful and number ownership verified')
        return _('Error sending sms')


class TestAlexaHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        result = sickrage.app.notification_providers['alexa'].test_notify()
        if result:
            return _('Alexa notification successful')
        return _('Alexa notification failed')


class TestSlackHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        result = sickrage.app.notification_providers['slack'].test_notify()
        if result:
            return _('Slack message successful')
        return _('Slack message failed')


class TestDiscordHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        result = sickrage.app.notification_providers['discord'].test_notify()
        if result:
            return _('Discord message successful')
        return _('Discord message failed')


class TestKODIHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_hosts(self.get_argument('host'))
        username = self.get_argument('username')
        password = self.get_argument('password')

        final_result = ''
        for curHost in [x.strip() for x in host.split(",")]:
            cur_result = sickrage.app.notification_providers['kodi'].test_notify(unquote_plus(curHost), username, password)
            if len(cur_result.split(":")) > 2 and 'OK' in cur_result.split(":")[2]:
                final_result += _('Test KODI notice sent successfully to ') + unquote_plus(curHost)
            else:
                final_result += _('Test KODI notice failed to ') + unquote_plus(curHost)
            final_result += "<br>\n"

        return final_result


class TestPMCHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_hosts(self.get_argument('host'))
        username = self.get_argument('username')
        password = self.get_argument('password', None)

        if password and set('*') == set(password):
            password = sickrage.app.config.plex.client_password

        final_result = ''
        for curHost in [x.strip() for x in host.split(',')]:
            cur_result = sickrage.app.notification_providers['plex'].test_notify_pmc(unquote_plus(curHost), username,
                                                                                     password)
            if len(cur_result.split(':')) > 2 and 'OK' in cur_result.split(':')[2]:
                final_result += _('Successful test notice sent to Plex client ... ') + unquote_plus(curHost)
            else:
                final_result += _('Test failed for Plex client ... ') + unquote_plus(curHost)
            final_result += '<br>' + '\n'

        sickrage.app.alerts.message(_('Tested Plex client(s): '),
                                    unquote_plus(host.replace(',', ', ')))

        return final_result


class TestPMSHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_hosts(self.get_argument('host'))
        username = self.get_argument('username')
        password = self.get_argument('password', None)
        plex_server_token = self.get_argument('plex_server_token')

        if password and set('*') == set(password):
            password = sickrage.app.config.plex.password

        final_result = ''

        cur_result = sickrage.app.notification_providers['plex'].test_notify_pms(unquote_plus(host), username, password,
                                                                                 plex_server_token)
        if cur_result is None:
            final_result += _('Successful test of Plex server(s) ... ') + \
                            unquote_plus(host.replace(',', ', '))
        elif cur_result is False:
            final_result += _('Test failed, No Plex Media Server host specified')
        else:
            final_result += _('Test failed for Plex server(s) ... ') + \
                            unquote_plus(str(cur_result).replace(',', ', '))
        final_result += '<br>' + '\n'

        sickrage.app.alerts.message(_('Tested Plex Media Server host(s): '),
                                    unquote_plus(host.replace(',', ', ')))

        return final_result


class TestLibnotifyHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        if sickrage.app.notification_providers['libnotify'].test_notify():
            return _('Tried sending desktop notification via libnotify')
        return sickrage.app.notification_providers['libnotify'].diagnose()


class TestEMBYHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'))
        emby_apikey = self.get_argument('emby_apikey')

        result = sickrage.app.notification_providers['emby'].test_notify(unquote_plus(host), emby_apikey)
        if result:
            return _('Test notice sent successfully to ') + unquote_plus(host)
        return _('Test notice failed to ') + unquote_plus(host)


class TestNMJHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'))
        database = self.get_argument('database')
        mount = self.get_argument('mount')

        result = sickrage.app.notification_providers['nmj'].test_notify(unquote_plus(host), database, mount)
        if result:
            return _('Successfully started the scan update')
        return _('Test failed to start the scan update')


class SettingsNMJHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'))

        result = sickrage.app.notification_providers['nmj'].notify_settings(unquote_plus(host))
        if result:
            return '{"message": "%(message)s %(host)s", "database": "%(database)s", "mount": "%(mount)s"}' % {
                "message": _('Got settings from'),
                "host": host, "database": sickrage.app.config.nmj.database,
                "mount": sickrage.app.config.nmj.mount
            }

        message = _('Failed! Make sure your Popcorn is on and NMJ is running. (see Log & Errors -> Debug for '
                    'detailed info)')

        return '{"message": {}, "database": "", "mount": ""}'.format(message)


class TestNMJv2Handler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'))

        result = sickrage.app.notification_providers['nmjv2'].test_notify(unquote_plus(host))
        if result:
            return _('Test notice sent successfully to ') + unquote_plus(host)
        return _('Test notice failed to ') + unquote_plus(host)


class SettingsNMJv2Handler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'))
        dbloc = self.get_argument('dbloc')
        instance = self.get_argument('instance')

        result = sickrage.app.notification_providers['nmjv2'].notify_settings(unquote_plus(host), dbloc, instance)

        if result:
            return f'{{"message": "NMJ Database found at: {host}", "database": "{sickrage.app.config.nmjv2.database}"}}'

        return f'{{"message": "Unable to find NMJ Database at location: {dbloc}. Is the right location selected and PCH running?", "database": ""}}'


class GetTraktTokenHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        trakt_pin = self.get_argument('trakt_pin')

        if TraktAPI().authenticate(trakt_pin):
            return _('Trakt Authorized')
        return _('Trakt Not Authorized!')


class TestTraktHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        username = self.get_argument('username')
        blacklist_name = self.get_argument('blacklist_name')

        return sickrage.app.notification_providers['trakt'].test_notify(username, blacklist_name)


class LoadShowNotifyListsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        data = {'_size': 0}
        for s in sorted(get_show_list(), key=lambda k: k.name):
            data[s.series_id] = {'id': s.series_id, 'name': s.name, 'list': s.notify_list}
            data['_size'] += 1
        return json_encode(data)


class SaveShowNotifyListHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        emails = self.get_argument('emails')

        try:
            show = find_show(int(show))
            show.notify_list = emails
        except Exception:
            return 'ERROR'


class TestEmailHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        host = clean_host(self.get_argument('host'))
        port = self.get_argument('port')
        smtp_from = self.get_argument('smtp_from')
        use_tls = self.get_argument('use_tls')
        user = self.get_argument('user')
        pwd = self.get_argument('pwd')
        to = self.get_argument('to')

        if sickrage.app.notification_providers['email'].test_notify(host, port, smtp_from, use_tls, user, pwd, to):
            return _('Test email sent successfully! Check inbox.')
        return _('ERROR: %s') % sickrage.app.notification_providers['email'].last_err


class TestNMAHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        nma_api = self.get_argument('nma_api')
        nma_priority = self.get_argument('nma_priority')

        result = sickrage.app.notification_providers['nma'].test_notify(nma_api, nma_priority)
        if result:
            return _('Test NMA notice sent successfully')
        return _('Test NMA notice failed')


class TestPushalotHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        authorization_token = self.get_argument('authorizationToken')

        result = sickrage.app.notification_providers['pushalot'].test_notify(authorization_token)
        if result:
            return _('Pushalot notification succeeded. Check your Pushalot clients to make sure it worked')
        return _('Error sending Pushalot notification')


class TestPushbulletHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        api = self.get_argument('api')

        result = sickrage.app.notification_providers['pushbullet'].test_notify(api)
        if result:
            return _('Pushbullet notification succeeded. Check your device to make sure it worked')
        return _('Error sending Pushbullet notification')


class GetPushbulletDevicesHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        api = self.get_argument('api')

        result = sickrage.app.notification_providers['pushbullet'].get_devices(api)
        if result:
            return result
        return _('Error getting Pushbullet devices')


class ServerStatusHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        tvdir_free = get_disk_space_usage(sickrage.app.config.general.tv_download_dir)
        root_dir = {}
        if sickrage.app.config.general.root_dirs:
            backend_pieces = sickrage.app.config.general.root_dirs.split('|')
            backend_dirs = backend_pieces[1:]
        else:
            backend_dirs = []

        if len(backend_dirs):
            for subject in backend_dirs:
                root_dir[subject] = get_disk_space_usage(subject)

        return self.render('home/server_status.mako',
                           title=_('Server Status'),
                           header=_('Server Status'),
                           topmenu='system',
                           tvdirFree=tvdir_free,
                           rootDir=root_dir,
                           controller='home',
                           action='server_status')


class ProviderStatusHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('home/provider_status.mako',
                           title=_('Provider Status'),
                           header=_('Provider Status'),
                           topmenu='system',
                           controller='home',
                           action='provider_status')


class ShutdownHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        pid = self.get_argument('pid')

        if str(pid) != str(sickrage.app.pid):
            return self.redirect("/{}/".format(sickrage.app.config.general.default_page.value))

        self._genericMessage(_("Shutting down"), _("SiCKRAGE is shutting down"))

        sickrage.app.shutdown()


class RestartHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        pid = self.get_argument('pid')
        force = self.get_argument('force', None)

        if str(pid) != str(sickrage.app.pid) and not force:
            return self.redirect("/{}/".format(sickrage.app.config.general.default_page.value))

        # clear current user to disable header and footer
        self.current_user = None

        sickrage.app.scheduler.add_job(
            sickrage.app.restart,
            DateTrigger(
                run_date=datetime.datetime.utcnow() + datetime.timedelta(seconds=5),
                timezone='utc'
            )
        )

        return self.render('home/restart.mako',
                           title="Home",
                           header="Restarting SiCKRAGE",
                           topmenu="system",
                           controller='home',
                           action="restart")


class UpdateCheckHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        pid = self.get_argument('pid')

        if str(pid) != str(sickrage.app.pid):
            return self.redirect("/{}/".format(sickrage.app.config.general.default_page.value))

        sickrage.app.alerts.message(_("Updater"), _('Checking for updates'))

        # check for new app updates
        if not sickrage.app.version_updater.check_for_update():
            sickrage.app.alerts.message(_("Updater"), _('No new updates available!'))

        return self.redirect(self.previous_url())


class UpdateHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        pid = self.get_argument('pid')

        if str(pid) != str(sickrage.app.pid):
            return self.redirect("/{}/".format(sickrage.app.config.general.default_page.value))

        sickrage.app.alerts.message(_("Updater"), _('Updating SiCKRAGE'))

        sickrage.app.version_updater.update(webui=True)

        return self.redirect(self.previous_url())


class VerifyPathHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        path = self.get_argument('path')

        if os.path.isfile(path):
            return _('Successfully found {path}'.format(path=path))
        return _('Failed to find {path}'.format(path=path))


class InstallRequirementsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        sickrage.app.alerts.message(_('Upgrading PIP'))
        if sickrage.app.version_updater.updater.upgrade_pip():
            sickrage.app.alerts.message(_('Upgraded PIP successfully!'))

            sickrage.app.alerts.message(_('Installing SiCKRAGE requirements'))
            if sickrage.app.version_updater.updater.install_requirements(sickrage.app.version_updater.updater.current_branch):
                sickrage.app.alerts.message(_('Installed SiCKRAGE requirements successfully!'))
            else:
                sickrage.app.alerts.message(_('Failed to install SiCKRAGE requirements'))
        else:
            sickrage.app.alerts.message(_('Failed to upgrade PIP'))

        return self.redirect(self.previous_url())


class BranchCheckoutHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        branch = self.get_argument('branch')

        if sickrage.app.version_updater.updater.current_branch != branch:
            sickrage.app.alerts.message(_('Checking out branch: '), branch)
            if sickrage.app.version_updater.updater.checkout_branch(branch):
                sickrage.app.alerts.message(_('Branch checkout successful, restarting: '), branch)
                return self.redirect(url_concat("/home/restart", {'pid': sickrage.app.pid}))
        else:
            sickrage.app.alerts.message(_('Already on branch: '), branch)

        return self.redirect(self.previous_url())


class DisplayShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        submenu = []

        show_obj = find_show(int(show))
        if not show_obj:
            return self._genericMessage(_("Error"), _("Show not in show list"))

        episode_objects = sorted(show_obj.episodes, key=lambda x: (x.season, x.episode), reverse=True)
        season_results = set()

        submenu.append({
            'title': _('Edit'),
            'path': '/manage/editShow?show=%d' % show_obj.series_id,
            'icon': 'fas fa-edit'
        })

        show_loc = show_obj.location

        show_message = ''

        if sickrage.app.show_queue.is_being_added(show_obj.series_id):
            show_message = _('This show is in the process of being downloaded - the info below is incomplete.')

        elif sickrage.app.show_queue.is_being_updated(show_obj.series_id):
            show_message = _('The information on this page is in the process of being updated.')

        elif sickrage.app.show_queue.is_being_refreshed(show_obj.series_id):
            show_message = _('The episodes below are currently being refreshed from disk')

        elif sickrage.app.show_queue.is_being_subtitled(show_obj.series_id):
            show_message = _('Currently downloading subtitles for this show')

        elif sickrage.app.show_queue.is_queued_to_refresh(show_obj.series_id):
            show_message = _('This show is queued to be refreshed.')

        elif sickrage.app.show_queue.is_queued_to_update(show_obj.series_id):
            show_message = _('This show is queued and awaiting an update.')

        elif sickrage.app.show_queue.is_queued_to_subtitle(show_obj.series_id):
            show_message = _('This show is queued and awaiting subtitles download.')

        if not sickrage.app.show_queue.is_being_added(show_obj.series_id):
            if not sickrage.app.show_queue.is_being_updated(show_obj.series_id):
                if show_obj.paused:
                    submenu.append({
                        'title': _('Resume'),
                        'path': '/home/togglePause?show=%d' % show_obj.series_id,
                        'icon': 'fas fa-play'
                    })
                else:
                    submenu.append({
                        'title': _('Pause'),
                        'path': '/home/togglePause?show=%d' % show_obj.series_id,
                        'icon': 'fas fa-pause'
                    })

                submenu.append({
                    'title': _('Remove'),
                    'path': '/home/deleteShow?show=%d' % show_obj.series_id,
                    'class': 'removeshow',
                    'confirm': True,
                    'icon': 'fas fa-trash'
                })

                submenu.append({
                    'title': _('Re-scan files'),
                    'path': '/home/refreshShow?show=%d' % show_obj.series_id,
                    'icon': 'fas fa-compass'
                })

                submenu.append({
                    'title': _('Full Update'),
                    'path': '/home/updateShow?show=%d&amp;force=1' % show_obj.series_id,
                    'icon': 'fas fa-sync'
                })

                submenu.append({
                    'title': _('Update show in KODI'),
                    'path': '/home/updateKODI?show=%d' % show_obj.series_id,
                    'requires': self.have_kodi(),
                    'icon': 'fas fa-tv'
                })

                submenu.append({
                    'title': _('Update show in Emby'),
                    'path': '/home/updateEMBY?show=%d' % show_obj.series_id,
                    'requires': self.have_emby(),
                    'icon': 'fas fa-tv'
                })

                submenu.append({
                    'title': _('Preview Rename'),
                    'path': '/home/testRename?show=%d' % show_obj.series_id,
                    'icon': 'fas fa-tag'
                })

                if sickrage.app.config.subtitles.enable and show_obj.subtitles:
                    if not sickrage.app.show_queue.is_being_subtitled(show_obj.series_id):
                        submenu.append({
                            'title': _('Download Subtitles'),
                            'path': '/home/subtitleShow?show=%d' % show_obj.series_id,
                            'icon': 'fas fa-comment'
                        })

        ep_cats = {}
        ep_counts = {
            Overview.SKIPPED: 0,
            Overview.WANTED: 0,
            Overview.LOW_QUALITY: 0,
            Overview.GOOD: 0,
            Overview.UNAIRED: 0,
            Overview.SNATCHED: 0,
            Overview.SNATCHED_PROPER: 0,
            Overview.SNATCHED_BEST: 0,
            Overview.MISSED: 0,
        }

        for episode_object in episode_objects:
            season_results.add(episode_object.season)

            cur_ep_cat = episode_object.overview or -1

            if episode_object.airdate > datetime.date.min:
                today = datetime.datetime.now().replace(tzinfo=sickrage.app.tz).date()
                air_date = episode_object.airdate
                if air_date.year >= 1970 or show_obj.network:
                    air_date = SRDateTime(sickrage.app.tz_updater.parse_date_time(episode_object.airdate, show_obj.airs, show_obj.network),
                                          convert=True).dt.date()

                if cur_ep_cat == Overview.WANTED and air_date < today:
                    cur_ep_cat = Overview.MISSED

            if cur_ep_cat:
                ep_cats[str(episode_object.season) + "x" + str(episode_object.episode)] = cur_ep_cat
                ep_counts[cur_ep_cat] += 1

        if sickrage.app.config.anidb.split_home:
            shows, anime = [], []
            for show in get_show_list():
                if show.is_anime:
                    anime.append(show)
                else:
                    shows.append(show)

            sorted_show_lists = {
                "Shows": sorted(shows, key=lambda x: x.name.upper()),
                "Anime": sorted(anime, key=lambda x: x.name.upper())
            }
        else:
            sorted_show_lists = {
                "Shows": sorted(get_show_list(), key=lambda x: x.name.upper())
            }

        bwl = None
        if show_obj.is_anime:
            bwl = show_obj.release_groups

        # Insert most recent show
        for index, recentShow in enumerate(sickrage.app.shows_recent):
            if recentShow['series_id'] == show_obj.series_id:
                break
        else:
            sickrage.app.shows_recent.append({
                'series_id': show_obj.series_id,
                'name': show_obj.name,
            })

        return self.render('home/display_show.mako',
                           submenu=submenu,
                           showLoc=show_loc,
                           show_message=show_message,
                           show=show_obj,
                           episode_objects=episode_objects,
                           seasonResults=list(season_results),
                           sortedShowLists=sorted_show_lists,
                           bwl=bwl,
                           epCounts=ep_counts,
                           epCats=ep_cats,
                           scene_numbering=get_scene_numbering_for_show(show_obj.series_id, show_obj.series_provider_id),
                           xem_numbering=get_xem_numbering_for_show(show_obj.series_id, show_obj.series_provider_id),
                           scene_absolute_numbering=get_scene_absolute_numbering_for_show(show_obj.series_id, show_obj.series_provider_id),
                           xem_absolute_numbering=get_xem_absolute_numbering_for_show(show_obj.series_id, show_obj.series_provider_id),
                           title=show_obj.name,
                           controller='home',
                           action="display_show")

    def have_kodi(self):
        return sickrage.app.config.kodi.enable and sickrage.app.config.kodi.update_library

    def have_plex(self):
        return sickrage.app.config.plex.enable and sickrage.app.config.plex.update_library

    def have_emby(self):
        return sickrage.app.config.emby.enable


class TogglePauseHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        show_obj = find_show(int(show))

        if show_obj is None:
            return self._genericMessage(_("Error"), _("Unable to find the specified show"))

        show_obj.paused = not show_obj.paused
        show_obj.save()

        sickrage.app.alerts.message(
            _('%s has been %s') % (show_obj.name, (_('resumed'), _('paused'))[show_obj.paused]))

        return self.redirect("/home/displayShow?show=%i" % show_obj.series_id)


class DeleteShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        full = self.get_argument('full', None)

        show_obj = find_show(int(show))

        if show_obj is None:
            return self._genericMessage(_("Error"), _("Unable to find the specified show"))

        try:
            sickrage.app.show_queue.remove_show(show_obj.series_id, show_obj.series_provider_id, bool(full))
            sickrage.app.alerts.message(
                _('%s has been %s %s') %
                (
                    show_obj.name,
                    (_('deleted'), _('trashed'))[bool(sickrage.app.config.general.trash_remove_show)],
                    (_('(media untouched)'), _('(with all related media)'))[bool(full)]
                )
            )
        except CantRemoveShowException as e:
            sickrage.app.alerts.error(_('Unable to delete this show.'), str(e))

        sleep(sickrage.app.config.general.cpu_preset.value)

        # Don't redirect to the default page, so the user can confirm that the show was deleted
        return self.redirect('/home/')


class RefreshShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        show_obj = find_show(int(show))

        if show_obj is None:
            return self._genericMessage(_("Error"), _("Unable to find the specified show"))

        try:
            sickrage.app.show_queue.refresh_show(show_obj.series_id, show_obj.series_provider_id, True)
        except CantRefreshShowException as e:
            sickrage.app.alerts.error(_('Unable to refresh this show.'), str(e))

        sleep(sickrage.app.config.general.cpu_preset.value)

        return self.redirect("/home/displayShow?show=" + str(show_obj.series_id))


class UpdateShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        force = self.get_argument('force', None)

        show_obj = find_show(int(show))

        if show_obj is None:
            return self._genericMessage(_("Error"), _("Unable to find the specified show"))

        # force the update
        try:
            sickrage.app.show_queue.update_show(show_obj.series_id, show_obj.series_provider_id, force=bool(force))
        except CantUpdateShowException as e:
            sickrage.app.alerts.error(_("Unable to update this show."), str(e))

        # just give it some time
        sleep(sickrage.app.config.general.cpu_preset.value)

        return self.redirect("/home/displayShow?show=" + str(show_obj.series_id))


class SubtitleShowHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        show_obj = find_show(int(show))

        if show_obj is None:
            return self._genericMessage(_("Error"), _("Unable to find the specified show"))

        # search and download subtitles
        sickrage.app.show_queue.download_subtitles(show_obj.series_id, show_obj.series_provider_id)

        sleep(sickrage.app.config.general.cpu_preset.value)

        return self.redirect("/home/displayShow?show=" + str(show_obj.series_id))


class UpdateKODIHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        show_name = None

        show_obj = find_show(int(show))
        if show_obj:
            show_name = quote_plus(show_obj.name.encode())

        if show_name:
            if sickrage.app.config.kodi.update_only_first:
                host = sickrage.app.config.kodi.host.split(",")[0].strip()
            else:
                host = sickrage.app.config.kodi.host

            if sickrage.app.notification_providers['kodi'].update_library(showName=show_name):
                sickrage.app.alerts.message(_("Library update command sent to KODI host(s): ") + host)
            else:
                sickrage.app.alerts.error(_("Unable to contact one or more KODI host(s): ") + host)

        if show_obj:
            return self.redirect('/home/displayShow?show=' + str(show_obj.series_id))
        else:
            return self.redirect('/home/')


class UpdatePLEXHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        if not sickrage.app.notification_providers['plex'].update_library():
            sickrage.app.alerts.message(
                _("Library update command sent to Plex Media Server host: ") +
                sickrage.app.config.plex.server_host)
        else:
            sickrage.app.alerts.error(
                _("Unable to contact Plex Media Server host: ") +
                sickrage.app.config.plex.server_host)
        return self.redirect('/home/')


class UpdateEMBYHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        show_obj = find_show(int(show))

        if show_obj:
            if sickrage.app.notification_providers['emby'].update_library(show_obj):
                sickrage.app.alerts.message(
                    _("Library update command sent to Emby host: ") + sickrage.app.config.emby.host)
            else:
                sickrage.app.alerts.error(
                    _("Unable to contact Emby host: ") + sickrage.app.config.emby.host)

            return self.redirect('/home/displayShow?show=' + str(show_obj.series_id))
        else:
            return self.redirect('/home/')


class SyncTraktHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        sickrage.app.log.info("Syncing Trakt with SiCKRAGE")
        sickrage.app.alerts.message(_('Syncing Trakt with SiCKRAGE'))

        job = sickrage.app.scheduler.get_job(sickrage.app.trakt_searcher.name)
        if job:
            job.modify(next_run_time=datetime.datetime.utcnow(), kwargs={'force': True})

        return self.redirect("/home/")


class DeleteEpisodeHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        eps = self.get_argument('eps')
        direct = self.get_argument('direct', None)

        show_obj = find_show(int(show))

        if not show_obj:
            err_msg = _("Error", "Show not in show list")
            if direct:
                sickrage.app.alerts.error(_('Error'), err_msg)
                return json_encode({'result': 'error'})
            else:
                return self._genericMessage(_("Error"), err_msg)

        if eps:
            for curEp in eps.split('|'):
                if not curEp:
                    sickrage.app.log.debug("curEp was empty when trying to deleteEpisode")

                sickrage.app.log.debug("Attempting to delete episode " + curEp)

                ep_info = curEp.split('x')
                if not len(ep_info):
                    continue

                season = int(ep_info[0])
                episode = int(ep_info[1])

                try:
                    if not show_obj.delete_episode(season, episode, full=True):
                        return self._genericMessage(_("Error"), _("Episode couldn't be retrieved"))
                except EpisodeDeletedException:
                    pass

        if direct:
            return json_encode({'result': 'success'})
        else:
            return self.redirect("/home/displayShow?show=" + show)


class TestRenameHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        show_object = find_show(int(show))

        if show_object is None:
            return self._genericMessage(_("Error"), _("Show not in show list"))

        if not os.path.isdir(show_object.location):
            return self._genericMessage(_("Error"), _("Can't rename episodes when the show dir is missing."))

        episode_objects = []

        for cur_ep_obj in (x for x in show_object.episodes if x.location):
            if cur_ep_obj.location:
                if cur_ep_obj.related_episodes:
                    for cur_related_ep in cur_ep_obj.related_episodes + [cur_ep_obj]:
                        if cur_related_ep in episode_objects:
                            break
                        episode_objects.append(cur_ep_obj)
                else:
                    episode_objects.append(cur_ep_obj)

        if episode_objects:
            episode_objects.reverse()

        submenu = [
            {'title': _('Edit'), 'path': '/manage/editShow?show=%d' % show_object.series_id,
             'icon': 'fas fa-edit'}]

        return self.render('home/test_renaming.mako',
                           submenu=submenu,
                           episode_objects=episode_objects,
                           show=show_object,
                           title=_('Preview Rename'),
                           header=_('Preview Rename'),
                           controller='home',
                           action="test_renaming")


class DoRenameHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        eps = self.get_argument('eps')

        tv_show = find_show(int(show))
        if tv_show is None:
            err_msg = _("Show not in show list")
            return self._genericMessage(_("Error"), err_msg)

        if not os.path.isdir(tv_show.location):
            return self._genericMessage(_("Error"), _("Can't rename episodes when the show dir is missing."))

        if eps is None:
            return self.redirect("/home/displayShow?show=" + show)

        for curEp in eps.split('|'):
            ep_info = curEp.split('x')
            root_ep_season = int(ep_info[0])
            root_ep_episode = int(ep_info[1])

            try:
                root_ep_obj = tv_show.get_episode(season=root_ep_season, episode=root_ep_episode)
            except EpisodeNotFoundException:
                sickrage.app.log.warning("Unable to find an episode for " + curEp + ", skipping")
                continue

            root_ep_obj.rename()

        return self.redirect("/home/displayShow?show=" + show)


class SearchEpisodeHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        series_provider_id = self.get_argument('seriesProviderID')
        season = self.get_argument('season')
        episode = self.get_argument('episode')
        down_cur_quality = self.get_argument('downCurQuality')

        # make a queue item for it and put it on the queue
        ep_queue_item = ManualSearchTask(int(show), SeriesProviderID[series_provider_id], int(season), int(episode), bool(int(down_cur_quality)))

        sickrage.app.search_queue.put(ep_queue_item)
        if not all([ep_queue_item.started, ep_queue_item.success]):
            return json_encode({'result': 'success'})

        return json_encode({'result': 'failure'})


class GetManualSearchStatusHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')

        episodes = []

        for task_id, task_items in sickrage.app.search_queue.TASK_HISTORY.copy().items():
            search_task = sickrage.app.search_queue.fetch_task(task_id)
            if not search_task:
                # Finished Manual Searches
                episodes += self.get_episodes(int(show), task_items['season'], task_items['episode'], 'Finished')
                if task_id in sickrage.app.search_queue.TASK_HISTORY:
                    del sickrage.app.search_queue.TASK_HISTORY[task_id]

            if isinstance(search_task, (ManualSearchTask, FailedSearchTask)):
                if search_task.status == TaskStatus.QUEUED:
                    # Queued Manual Searches
                    episodes += self.get_episodes(int(show), task_items['season'], task_items['episode'], 'Queued')
                elif search_task.status == TaskStatus.STARTED:
                    # Running Manual Searches
                    episodes += self.get_episodes(int(show), task_items['season'], task_items['episode'], 'Searching')

        return json_encode({'episodes': episodes})

    def get_episodes(self, series_id, season, episode, search_status):
        results = []

        if not series_id:
            return results

        show_object = find_show(series_id)
        if not show_object:
            return results

        try:
            episode_object = show_object.get_episode(season, episode)
        except EpisodeNotFoundException:
            return results

        results.append({'show': series_id,
                        'season': episode_object.season,
                        'episode': episode_object.episode,
                        'searchstatus': search_status,
                        'status': episode_object.status.display_name,
                        'quality': self.get_quality_class(episode_object.status),
                        'overview': episode_object.overview.css_name})

        return results

    def get_quality_class(self, status):
        __, ep_quality = Quality.split_composite_status(status)
        if ep_quality in Qualities:
            quality_class = ep_quality.css_name
        else:
            quality_class = Qualities.UNKNOWN.css_name

        return quality_class


class SearchEpisodeSubtitlesHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        season = self.get_argument('season')
        episode = self.get_argument('episode')

        tv_show = find_show(int(show))
        if tv_show is None:
            return _("Invalid show paramaters")

        try:
            tv_episode = tv_show.get_episode(int(season), int(episode))
            subtitles = tv_episode.download_subtitles()

            if subtitles:
                languages = [Subtitles().name_from_code(subtitle) for subtitle in subtitles]
                status = _('New subtitles downloaded: %s') % ', '.join([lang for lang in languages])
            else:
                status = _('No subtitles downloaded')

            sickrage.app.alerts.message(tv_show.name, status)
            return json_encode({'result': status, 'subtitles': ','.join(tv_episode.subtitles)})
        except (EpisodeNotFoundException, MultipleEpisodesInDatabaseException):
            return json_encode({'result': _("Episode couldn't be retrieved")})
        except Exception:
            return json_encode({'result': 'failure'})


class SetSceneNumberingHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        series_provider_id = self.get_argument('series_provider_id')
        for_season = self.get_argument('forSeason', '')
        for_episode = self.get_argument('forEpisode', '')
        for_absolute = self.get_argument('forAbsolute', '')
        scene_season = self.get_argument('sceneSeason', '')
        scene_episode = self.get_argument('sceneEpisode', '')
        scene_absolute = self.get_argument('sceneAbsolute', '')

        # sanitize:
        if for_season in ['null', '']:
            for_season = None
        if for_episode in ['null', '']:
            for_episode = None
        if for_absolute in ['null', '']:
            for_absolute = None
        if scene_season in ['null', '']:
            scene_season = None
        if scene_episode in ['null', '']:
            scene_episode = None
        if scene_absolute in ['null', '']:
            scene_absolute = None

        show_obj = find_show(int(show))
        if show_obj.is_anime:
            result = {
                'success': True,
                'forAbsolute': for_absolute,
                'sceneAbsolute': 0
            }
        else:
            result = {
                'success': True,
                'forSeason': for_season,
                'forEpisode': for_episode,
                'sceneSeason': 0,
                'sceneEpisode': 0
            }

        try:
            if for_absolute is not None:
                show_obj.get_episode(absolute_number=for_absolute)

                show = int(show)
                series_provider_id = SeriesProviderID[series_provider_id]
                for_absolute = int(for_absolute)
                if scene_absolute is not None:
                    scene_absolute = int(scene_absolute)

                if set_scene_numbering(show, series_provider_id, absolute_number=for_absolute, scene_absolute=scene_absolute):
                    sickrage.app.log.debug("setAbsoluteSceneNumbering for %s from %s to %s" % (show, for_absolute, scene_absolute))
                    if scene_absolute is not None:
                        result['sceneAbsolute'] = get_scene_absolute_numbering(show, series_provider_id, for_absolute)
                else:
                    result['errorMessage'] = _("Another episode already has the same scene absolute numbering")
                    result['success'] = False
            else:
                show_obj.get_episode(season=for_season, episode=for_episode)

                show = int(show)
                series_provider_id = SeriesProviderID[series_provider_id]
                for_season = int(for_season)
                for_episode = int(for_episode)
                if scene_season is not None:
                    scene_season = int(scene_season)
                if scene_episode is not None:
                    scene_episode = int(scene_episode)

                if set_scene_numbering(show, series_provider_id, season=for_season, episode=for_episode, scene_season=scene_season,
                                       scene_episode=scene_episode):
                    sickrage.app.log.debug(
                        "setEpisodeSceneNumbering for %s from %sx%s to %sx%s" % (show, for_season, for_episode, scene_season, scene_episode))
                    if scene_season is not None and scene_episode is not None:
                        result['sceneSeason'], result['sceneEpisode'] = get_scene_numbering(show, series_provider_id, for_season, for_episode)
                else:
                    result['errorMessage'] = _("Another episode already has the same scene numbering")
                    result['success'] = False
        except EpisodeNotFoundException:
            result['errorMessage'] = _("Episode couldn't be retrieved")
            result['success'] = False

        return json_encode(result)


class RetryEpisodeHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show = self.get_argument('show')
        series_provider_id = self.get_argument('seriesProviderID')
        season = self.get_argument('season')
        episode = self.get_argument('episode')
        down_cur_quality = self.get_argument('downCurQuality')

        # retrieve the episode object and fail if we can't get one
        # make a queue item for it and put it on the queue
        ep_queue_item = FailedSearchTask(int(show), SeriesProviderID[series_provider_id], int(season), int(episode), bool(int(down_cur_quality)))

        sickrage.app.search_queue.put(ep_queue_item)
        if not all([ep_queue_item.started, ep_queue_item.success]):
            return json_encode({'result': 'success'})

        return json_encode({'result': 'failure'})


class FetchReleasegroupsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        show_name = self.get_argument('show_name')

        sickrage.app.log.info('ReleaseGroups: {}'.format(show_name))

        try:
            groups = get_release_groups_for_anime(show_name)
            sickrage.app.log.info('ReleaseGroups: {}'.format(groups))
        except AnidbAdbaConnectionException as e:
            sickrage.app.log.debug('Unable to get ReleaseGroups: {}'.format(e))
        else:
            return json_encode({'result': 'success', 'groups': groups})

        return json_encode({'result': 'failure'})



import collections
import errno
import time
import traceback
from urllib.parse import urljoin

import oauthlib.oauth2
import requests
import requests.exceptions
from jose import ExpiredSignatureError
from keycloak.exceptions import KeycloakClientError
from requests_oauthlib import OAuth2Session

import sickrage
from sickrage.core.api.exceptions import APIError


class API(object):
    def __init__(self):
        self.name = 'SR-API'
        self.api_base = 'https://www.sickrage.ca/api/'
        self.api_version = 'v6'
        self._token = {}

    @property
    def imdb(self):
        return self.IMDbAPI(self)

    @property
    def server(self):
        return self.ServerAPI(self)

    @property
    def search_provider(self):
        return self.SearchProviderAPI(self)

    @property
    def series_provider(self):
        return self.SeriesProviderAPI(self)

    @property
    def announcement(self):
        return self.AnnouncementsAPI(self)

    @property
    def google(self):
        return self.GoogleDriveAPI(self)

    @property
    def torrent(self):
        return self.TorrentAPI(self)

    @property
    def scene_exceptions(self):
        return self.SceneExceptions(self)

    @property
    def alexa(self):
        return self.AlexaAPI(self)

    @property
    def session(self):
        if not self.token_url:
            return

        return OAuth2Session(
            token=self.token,
            auto_refresh_kwargs={'client_id': sickrage.app.auth_server.client_id},
            auto_refresh_url=self.token_url,
            token_updater=self.token_updater
        )

    @property
    def token(self):
        if not self._token:
            self.login()
        elif self.token_time_remaining < (int(self._token.get('expires_in')) / 2):
            self.refresh_token()

        return self._token

    @property
    def token_expiration(self):
        try:
            if not self._token:
                return time.time()

            certs = sickrage.app.auth_server.certs()
            if not certs:
                return time.time()

            decoded_token = sickrage.app.auth_server.decode_token(self._token.get('access_token'), certs)
            return decoded_token.get('exp', time.time())
        except ExpiredSignatureError:
            return time.time()

    @property
    def token_time_remaining(self):
        return max(self.token_expiration - time.time(), 0)

    @property
    def token_is_expired(self):
        return self.token_expiration <= time.time()

    @property
    def token_url(self):
        return sickrage.app.auth_server.get_url('token_endpoint')

    @property
    def health(self):
        for i in range(3):
            try:
                health = requests.get(urljoin(self.api_base, "health"), verify=True, timeout=30).ok
            except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
                pass
            else:
                break
        else:
            health = False

        if not health:
            sickrage.app.log.debug("SiCKRAGE API is currently unreachable")
            return False

        return True

    @property
    def userinfo(self):
        return self.request('GET', 'userinfo')

    def token_updater(self, value):
        self._token = value

    def login(self):
        if not self.health:
            return False

        if not self.token_url:
            return False

        session = requests.session()

        data = {
            'client_id': sickrage.app.auth_server.client_id,
            'grant_type': 'password',
            'apikey': sickrage.app.config.general.sso_api_key
        }

        try:
            resp = session.post(self.token_url, data)
            resp.raise_for_status()
            self._token = resp.json()
        except requests.exceptions.RequestException:
            return False

        return True

    def logout(self):
        if self._token:
            try:
                sickrage.app.auth_server.logout(self._token.get('refresh_token'))
            except KeycloakClientError:
                pass

    def refresh_token(self):
        retries = 3

        for i in range(retries):
            try:
                if not self._token:
                    return self.login()

                self._token = sickrage.app.auth_server.refresh_token(self._token.get('refresh_token'))
            except KeycloakClientError:
                return self.login()
            except (requests.exceptions.ReadTimeout, requests.exceptions.ConnectionError):
                if i > retries:
                    return False
                time.sleep(0.2)
                continue

            return True

    def allowed_usernames(self):
        return self.request('GET', 'allowed-usernames')

    def download_privatekey(self):
        return self.request('GET', 'server/config/private-key')

    def upload_privatekey(self, privatekey):
        return self.request('POST', 'server/config/private-key', data=dict({'privatekey': privatekey}))

    def network_timezones(self):
        return self.request('GET', 'network-timezones')

    def request(self, method, url, timeout=120, **kwargs):
        if not self.session:
            return

        url = urljoin(self.api_base, "/".join([self.api_version, url]))

        for i in range(5):
            resp = None

            try:
                if not self.health:
                    if i > 3:
                        return None
                    continue

                resp = self.session.request(method, url, timeout=timeout, verify=False, hooks={'response': self.throttle_hook}, **kwargs)

                resp.raise_for_status()
                if resp.status_code == 204:
                    return resp.ok

                try:
                    return resp.json()
                except ValueError:
                    return resp.content
            except (oauthlib.oauth2.TokenExpiredError, oauthlib.oauth2.InvalidGrantError):
                self.refresh_token()
                time.sleep(1)
            except (oauthlib.oauth2.InvalidClientIdError, oauthlib.oauth2.MissingTokenError) as e:
                self.refresh_token()
                time.sleep(1)
            except requests.exceptions.ReadTimeout as e:
                if i > 3:
                    sickrage.app.log.debug(f'Error connecting to url {url} Error: {e}')
                    return resp or e.response

                timeout += timeout
                time.sleep(1)
            except requests.exceptions.HTTPError as e:
                status_code = e.response.status_code
                error_message = e.response.text

                if status_code == 403 and "login-pf-page" in error_message:
                    self.refresh_token()
                    time.sleep(1)
                    continue

                if 'application/json' in e.response.headers.get('content-type', ''):
                    status_code = e.response.json().get('error', status_code)
                    error_message = e.response.json().get('message', error_message)
                    sickrage.app.log.debug(f'SiCKRAGE API response returned for url {url} Response: {error_message}, Code: {status_code}')
                else:
                    sickrage.app.log.debug(f'The response returned a non-200 response while requesting url {url} Error: {e!r}')

                return resp or e.response
            except requests.exceptions.ConnectionError as e:
                if i > 3:
                    sickrage.app.log.debug(f'Error connecting to url {url} Error: {e}')
                    return resp or e.response

                time.sleep(1)
            except requests.exceptions.RequestException as e:
                sickrage.app.log.debug(f'Error requesting url {url} Error: {e}')
                return resp or e.response
            except Exception as e:
                if (isinstance(e, collections.Iterable) and 'ECONNRESET' in e) or (getattr(e, 'errno', None) == errno.ECONNRESET):
                    sickrage.app.log.warning(f'Connection reset by peer accessing url {url} Error: {e}')
                else:
                    sickrage.app.log.info(f'Unknown exception in url {url} Error: {e}')
                    sickrage.app.log.debug(traceback.format_exc())

                return None

    @staticmethod
    def throttle_hook(response, **kwargs):
        if "X-RateLimit-Remaining" in response.headers:
            remaining = int(response.headers["X-RateLimit-Remaining"])
            if remaining == 1:
                sickrage.app.log.debug("Throttling SiCKRAGE API Calls... Sleeping for 60 secs...\n")
                time.sleep(60)

    class ServerAPI:
        def __init__(self, api):
            self.api = api

        def register_server(self, ip_addresses, web_protocol, web_port, web_root, server_version):
            data = {
                'ip-addresses': ip_addresses,
                'web-protocol': web_protocol,
                'web-port': web_port,
                'web-root': web_root,
                'server-version': server_version
            }

            return self.api.request('POST', 'server', data=data)

        def unregister_server(self, server_id):
            data = {
                'server-id': server_id
            }

            return self.api.request('DELETE', 'server', data=data)

        def update_server(self, server_id, ip_addresses, web_protocol, web_port, web_root, server_version):
            data = {
                'server-id': server_id,
                'ip-addresses': ip_addresses,
                'web-protocol': web_protocol,
                'web-port': web_port,
                'web-root': web_root,
                'server-version': server_version
            }

            return self.api.request('PUT', 'server', data=data)

        def get_status(self, server_id):
            return self.api.request('GET', f'server/{server_id}/status')

        def get_server_certificate(self, server_id):
            return self.api.request('GET', f'server/{server_id}/certificate')

        def declare_amqp_queue(self, server_id):
            return self.api.request('GET', f'server/{server_id}/declare-amqp-queue')

        def upload_config(self, server_id, pkey_sig, config):
            data = {
                'server-id': server_id,
                'pkey-sig': pkey_sig,
                'config': config
            }
            return self.api.request('POST', f'server/{server_id}/config', data=data)

        def download_config(self, server_id, pkey_sig):
            data = {
                'pkey-sig': pkey_sig
            }

            return self.api.request('GET', f'server/{server_id}/config', json=data)['config']

    class AnnouncementsAPI:
        def __init__(self, api):
            self.api = api

        def get_announcements(self):
            return self.api.request('GET', 'announcements')

    class SearchProviderAPI:
        def __init__(self, api):
            self.api = api

        def get_url(self, provider):
            endpoint = f'provider/{provider}/url'
            return self.api.request('GET', endpoint)

        def get_status(self, provider):
            endpoint = f'provider/{provider}/status'
            return self.api.request('GET', endpoint)

        def get_search_result(self, provider, series_id, season, episode):
            endpoint = f'provider/{provider}/series-id/{series_id}/season/{season}/episode/{episode}'
            return self.api.request('GET', endpoint)

        def add_search_result(self, provider, data):
            return self.api.request('POST', f'provider/{provider}', json=data)

    class TorrentAPI:
        def __init__(self, api):
            self.api = api

        def get_trackers(self):
            endpoint = f'torrent/trackers'
            return self.api.request('GET', endpoint)

        def get_torrent(self, hash):
            endpoint = f'torrent/{hash}'
            return self.api.request('GET', endpoint)

        def add_torrent(self, url):
            return self.api.request('POST', 'torrent', json={'url': url})

    class IMDbAPI:
        def __init__(self, api):
            self.api = api

        def search_by_imdb_title(self, title):
            endpoint = f'imdb/search-by-title/{title}'
            return self.api.request('GET', endpoint)

        def search_by_imdb_id(self, imdb_id):
            endpoint = f'imdb/search-by-id/{imdb_id}'
            return self.api.request('GET', endpoint)

    class GoogleDriveAPI:
        def __init__(self, api):
            self.api = api

        def is_connected(self):
            endpoint = 'google-drive/is-connected'
            return self.api.request('GET', endpoint)

        def upload(self, file, folder):
            endpoint = 'google-drive/upload'
            return self.api.request('POST', endpoint, files={'file': open(file, 'rb')}, params={'folder': folder})

        def download(self, id):
            endpoint = f'google-drive/download/{id}'
            return self.api.request('GET', endpoint)

        def delete(self, id):
            endpoint = f'google-drive/delete/{id}'
            return self.api.request('GET', endpoint)

        def search_files(self, id, term):
            endpoint = f'google-drive/search-files/{id}/{term}'
            return self.api.request('GET', endpoint)

        def list_files(self, id):
            endpoint = f'google-drive/list-files/{id}'
            return self.api.request('GET', endpoint)

        def clear_folder(self, id):
            endpoint = f'google-drive/clear-folder/{id}'
            return self.api.request('GET', endpoint)

    class SceneExceptions:
        def __init__(self, api):
            self.api = api

        def get(self, *args, **kwargs):
            endpoint = 'scene-exceptions'
            return self.api.request('GET', endpoint)

        def search_by_id(self, series_id):
            endpoint = f'scene-exceptions/search-by-id/{series_id}'
            return self.api.request('GET', endpoint)

    class AlexaAPI:
        def __init__(self, api):
            self.api = api

        def send_notification(self, message):
            return self.api.request('POST', 'alexa/notification', json={'message': message})

    class SeriesProviderAPI:
        def __init__(self, api):
            self.api = api

        def search(self, provider, query, language='eng'):
            endpoint = f'series-provider/{provider}/search/{query}/{language}'
            return self.api.request('GET', endpoint)

        def search_by_id(self, provider, remote_id, language='eng'):
            endpoint = f'series-provider/{provider}/search-by-id/{remote_id}/{language}'
            return self.api.request('GET', endpoint)

        def get_series_info(self, provider, series_id, language='eng'):
            endpoint = f'series-provider/{provider}/series/{series_id}/{language}'
            return self.api.request('GET', endpoint)

        def get_episodes_info(self, provider, series_id, season_type='default', language='eng'):
            endpoint = f'series-provider/{provider}/series/{series_id}/episodes/{season_type}/{language}'
            return self.api.request('GET', endpoint)

        def languages(self, provider):
            endpoint = f'series-provider/{provider}/languages'
            return self.api.request('GET', endpoint)

        def updates(self, provider, since):
            endpoint = f'series-provider/{provider}/updates/{since}'
            return self.api.request('GET', endpoint)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import threading
import time

from sqlalchemy import orm

import sickrage
from sickrage.core.databases.cache import CacheDB
from sickrage.core.enums import SeriesProviderID
from sickrage.core.exceptions import CantRefreshShowException, CantUpdateShowException
from sickrage.core.tv.show.helpers import get_show_list


class ShowUpdater(object):
    def __init__(self):
        self.name = "SHOWUPDATER"
        self.lock = threading.Lock()
        self.running = False

    def task(self, force=False):
        if self.running and not force:
            return

        try:
            self.running = True

            # set thread name
            threading.current_thread().name = self.name

            session = sickrage.app.cache_db.session()

            update_timestamp = int(time.mktime(datetime.datetime.now().timetuple()))

            try:
                dbData = session.query(CacheDB.LastUpdate).filter_by(provider='theTVDB').one()
                last_update = int(dbData.time)
            except orm.exc.NoResultFound:
                last_update = update_timestamp
                dbData = CacheDB.LastUpdate(**{
                    'provider': 'theTVDB',
                    'time': 0
                })
                session.add(dbData)
            finally:
                session.commit()

            # get list of updated series from a series provider
            updated_shows = set()
            for series_provider_id in SeriesProviderID:
                resp = sickrage.app.series_providers[series_provider_id].updates(last_update)
                if resp:
                    for series in resp:
                        updated_shows.add(series['id'])

            # start update process
            pi_list = []
            for show_obj in get_show_list():
                # if show_obj.paused:
                #     sickrage.app.log.info('Show update skipped, show: {} is paused.'.format(show_obj.name))
                #     continue

                if show_obj.status == 'Ended':
                    if not sickrage.app.config.general.show_update_stale:
                        sickrage.app.log.info('Show update skipped, show: {} status is ended.'.format(show_obj.name))
                        continue
                    elif not (datetime.datetime.now() - show_obj.last_update).days >= 90:
                        sickrage.app.log.info('Show update skipped, show: {} status is ended and recently updated.'.format(show_obj.name))
                        continue

                try:
                    if show_obj.series_id in updated_shows:
                        pi_list.append(sickrage.app.show_queue.refresh_show(show_obj.series_id, show_obj.series_provider_id, force=False))
                    elif (datetime.datetime.now() - show_obj.last_update).days >= 7:
                        pi_list.append(sickrage.app.show_queue.update_show(show_obj.series_id, show_obj.series_provider_id, force=False))
                except (CantUpdateShowException, CantRefreshShowException) as e:
                    sickrage.app.log.debug("Automatic update failed: {}".format(e))

            # ProgressIndicators.setIndicator('dailyShowUpdates', QueueProgressIndicator("Daily Show Updates", pi_list))

            dbData.time = update_timestamp
            session.commit()
        finally:
            self.running = False



from urllib.parse import urljoin

from markdown2 import _slugify as slugify

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import try_int, convert_size, validate_url, bs4_parser
from sickrage.search_providers import TorrentProvider


class MagnetDLProvider(TorrentProvider):
    def __init__(self):
        super().__init__("MagnetDL", "https://www.magnetdl.com", False)

        # custom settings
        self.custom_settings = {
            'custom_url': '',
            'confirmed': False,
            'minseed': 0,
            'minleech': 0
        }

        self.cache = TVCache(self)

    @property
    def urls(self):
        return {
            'rss': f'{self.url}/download/tv/age/desc/'
        }

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: {mode}".format(mode=mode))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    sickrage.app.log.debug("Search String: {search_string}".format(search_string=search_string))
                    search = slugify(search_string)
                    search_url = urljoin(self.url, "{}/{}/".format(search[0], search))
                else:
                    search_url = self.urls["rss"]

                if self.custom_settings['custom_url']:
                    if not validate_url(self.custom_settings['custom_url']):
                        sickrage.app.log.warning("Invalid custom url: {0}".format(self.custom_settings['custom_url']))
                        return results
                    search_url = urljoin(self.custom_settings['custom_url'], search_url.split(self.url)[1])

                resp = self.session.get(search_url, headers={"Accept": "application/html"})
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            torrent_table = html.find("table", class_="download")
            torrent_body = torrent_table.find("tbody") if torrent_table else []
            torrent_rows = torrent_body("tr") if torrent_body else []

            # Continue only if at least one Release is found
            if not torrent_rows:
                sickrage.app.log.debug("Data returned from provider does not contain any torrents")
                return results

            labels = [x.get_text(strip=True) for x in torrent_table.find("thead").find("tr")("th")]

            # Skip column headers
            for result in torrent_rows[0:-1:2]:
                try:
                    if len(result("td")) < len(labels):
                        continue

                    title = result.find("td", class_="n").find("a")["title"]
                    magnet = result.find("td", class_="m").find("a")["href"]
                    seeders = try_int(result.find("td", class_="s").get_text(strip=True))
                    leechers = try_int(result.find("td", class_="l").get_text(strip=True))
                    size = convert_size(result("td")[labels.index("Size")].get_text(strip=True) or "", -1)

                    if not all([title, magnet]):
                        continue

                    results += [
                        {"title": title, "link": magnet, "size": size, "seeders": seeders, "leechers": leechers}
                    ]

                    if mode != "RSS":
                        sickrage.app.log.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))
                except Exception as error:
                    sickrage.app.log.debug(f"Failed parsing provider. Traceback: {error}")
                    continue

        return results



"""Initial migration

Revision ID: 11
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import json
import os
from json import JSONDecodeError

import sqlalchemy as sa
from alembic import op
from keycloak.exceptions import KeycloakClientError
from sqlalchemy import orm, inspect

import sickrage

# revision identifiers, used by Alembic.
from sickrage.core import ConfigDB

revision = '11'
down_revision = '10'


def upgrade():
    pass


def downgrade():
    # Operations to reverse the above upgrade go here.
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: network_timezone_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19network_timezone_v1.proto\x12\x10\x61pp.protobufs.v1\"A\n\x1cSavedNetworkTimezoneResponse\x12\x0f\n\x07network\x18\x01 \x01(\t\x12\x10\n\x08timezone\x18\x02 \x01(\t\"1\n\x1e\x44\x65letedNetworkTimezoneResponse\x12\x0f\n\x07network\x18\x01 \x01(\tb\x06proto3')



_SAVEDNETWORKTIMEZONERESPONSE = DESCRIPTOR.message_types_by_name['SavedNetworkTimezoneResponse']
_DELETEDNETWORKTIMEZONERESPONSE = DESCRIPTOR.message_types_by_name['DeletedNetworkTimezoneResponse']
SavedNetworkTimezoneResponse = _reflection.GeneratedProtocolMessageType('SavedNetworkTimezoneResponse', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDNETWORKTIMEZONERESPONSE,
  '__module__' : 'network_timezone_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.SavedNetworkTimezoneResponse)
  })
_sym_db.RegisterMessage(SavedNetworkTimezoneResponse)

DeletedNetworkTimezoneResponse = _reflection.GeneratedProtocolMessageType('DeletedNetworkTimezoneResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETEDNETWORKTIMEZONERESPONSE,
  '__module__' : 'network_timezone_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.DeletedNetworkTimezoneResponse)
  })
_sym_db.RegisterMessage(DeletedNetworkTimezoneResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SAVEDNETWORKTIMEZONERESPONSE._serialized_start=47
  _SAVEDNETWORKTIMEZONERESPONSE._serialized_end=112
  _DELETEDNETWORKTIMEZONERESPONSE._serialized_start=114
  _DELETEDNETWORKTIMEZONERESPONSE._serialized_end=163
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: search_provider_url_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1csearch_provider_url_v1.proto\x12\x10\x61pp.protobufs.v1\"K\n\x1eSavedSearchProviderUrlResponse\x12\x13\n\x0bprovider_id\x18\x01 \x01(\t\x12\x14\n\x0cprovider_url\x18\x02 \x01(\tb\x06proto3')



_SAVEDSEARCHPROVIDERURLRESPONSE = DESCRIPTOR.message_types_by_name['SavedSearchProviderUrlResponse']
SavedSearchProviderUrlResponse = _reflection.GeneratedProtocolMessageType('SavedSearchProviderUrlResponse', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDSEARCHPROVIDERURLRESPONSE,
  '__module__' : 'search_provider_url_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.SavedSearchProviderUrlResponse)
  })
_sym_db.RegisterMessage(SavedSearchProviderUrlResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SAVEDSEARCHPROVIDERURLRESPONSE._serialized_start=50
  _SAVEDSEARCHPROVIDERURLRESPONSE._serialized_end=125
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: announcement_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61nnouncement_v1.proto\x12\x10\x61pp.protobufs.v1\"m\n\x1b\x43reatedAnnouncementResponse\x12\r\n\x05\x61hash\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\r\n\x05image\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x61te\x18\x05 \x01(\t\",\n\x1b\x44\x65letedAnnouncementResponse\x12\r\n\x05\x61hash\x18\x01 \x01(\tb\x06proto3')



_CREATEDANNOUNCEMENTRESPONSE = DESCRIPTOR.message_types_by_name['CreatedAnnouncementResponse']
_DELETEDANNOUNCEMENTRESPONSE = DESCRIPTOR.message_types_by_name['DeletedAnnouncementResponse']
CreatedAnnouncementResponse = _reflection.GeneratedProtocolMessageType('CreatedAnnouncementResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDANNOUNCEMENTRESPONSE,
  '__module__' : 'announcement_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.CreatedAnnouncementResponse)
  })
_sym_db.RegisterMessage(CreatedAnnouncementResponse)

DeletedAnnouncementResponse = _reflection.GeneratedProtocolMessageType('DeletedAnnouncementResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETEDANNOUNCEMENTRESPONSE,
  '__module__' : 'announcement_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.DeletedAnnouncementResponse)
  })
_sym_db.RegisterMessage(DeletedAnnouncementResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CREATEDANNOUNCEMENTRESPONSE._serialized_start=43
  _CREATEDANNOUNCEMENTRESPONSE._serialized_end=152
  _DELETEDANNOUNCEMENTRESPONSE._serialized_start=154
  _DELETEDANNOUNCEMENTRESPONSE._serialized_end=198
# @@protoc_insertion_point(module_scope)



import sickrage
from sickrage.core.helpers import backup_app_data


class AutoBackup(object):
    def __init__(self, *args, **kwargs):
        self.name = "AUTO-BACKUP"

    def task(self):
        if not sickrage.app.config.general.auto_backup_enable:
            return

        sickrage.app.log.info("Performing automatic backup of SiCKRAGE")
        backup_app_data(sickrage.app.config.general.auto_backup_dir, backup_type="auto", keep_num=sickrage.app.config.general.auto_backup_keep_num)
        sickrage.app.log.info("Finished automatic backup of SiCKRAGE")


# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import datetime
import fnmatch
import os
import re
from functools import partial

import sickrage
from sickrage.core.common import EpisodeStatus, countryList
from sickrage.core.enums import SearchFormat
from sickrage.core.helpers import sanitize_scene_name, strip_accents
from sickrage.core.tv.show.helpers import find_show

resultFilters = [
    "sub(bed|ed|pack|s)",
    "(dir|sub|nfo)fix",
    "(?<!shomin.)sample",
    "(dvd)?extras",
    "dub(bed)?"
]

if hasattr('General', 'ignored_subs_list') and sickrage.app.config.general.ignored_subs_list:
    resultFilters.append("(" + sickrage.app.config.general.ignored_subs_list.replace(",", "|") + ")sub(bed|ed|s)?")


def contains_at_least_one_word(name, words):
    """
    Filters out results based on filter_words

    name: name to check
    words : string of words separated by a ',' or list of words

    Returns: False if the name doesn't contain any word of words list, or the found word from the list.
    :return:
    :param name:
    :param words:
    :return:
    :rtype: unicode
    """
    if isinstance(words, str):
        words = words.split(',')
    items = [(re.compile(r'(^|[\W_])%s($|[\W_])' % re.escape(word.strip()), re.I), word.strip()) for word in words]
    for regexp, word in items:
        if regexp.search(name):
            return word
    return False


def filter_bad_releases(name, parse=True):
    """
    Filters out non-english and just all-around stupid releases by comparing them
    to the resultFilters contents.

    name: the release name to check

    Returns: True if the release name is OK, False if it's bad.
    :return:
    :param name:
    :param parse:
    :return:
    """

    from sickrage.core.nameparser import InvalidNameException, InvalidShowException, NameParser

    try:
        if parse:
            NameParser().parse(name)
    except InvalidNameException as e:
        sickrage.app.log.debug("{}".format(e))
        return False
    except InvalidShowException:
        pass

    # if any of the bad strings are in the name then say no
    ignore_words = list(resultFilters)
    if sickrage.app.config.general.ignore_words:
        ignore_words.extend(sickrage.app.config.general.ignore_words.split(','))

    word = contains_at_least_one_word(name, ignore_words)
    if word:
        sickrage.app.log.debug("Release: " + name + " contains " + word + ", ignoring it")
        return False

    # if any of the good strings aren't in the name then say no
    if sickrage.app.config.general.require_words:
        require_words = sickrage.app.config.general.require_words
        if not contains_at_least_one_word(name, require_words):
            sickrage.app.log.debug("Release: " + name + " doesn't contain any of " +
                                   ', '.join(set(require_words)) + ", ignoring it")
            return False

    return True


def scene_to_normal_show_names(name):
    """
        Takes a show name from a scene dirname and converts it to a more "human-readable" format.

    name: The show name to convert

    Returns: a list of all the possible "normal" names
    :return:
    :param name:
    """

    if not name:
        return []

    name_list = [name]

    # use both and and &
    new_name = re.sub(r'(?i)([\. ])and([\. ])', '\\1&\\2', name, re.I)
    if new_name not in name_list:
        name_list.append(new_name)

    results = []

    for cur_name in name_list:
        # add brackets around the year
        results.append(re.sub(r'(\D)(\d{4})$', '\\1(\\2)', cur_name))

        # add brackets around the country
        country_match_str = '|'.join(countryList.values())
        results.append(re.sub(r'(?i)([. _-])(' + country_match_str + ')$', '\\1(\\2)', cur_name))

    results += name_list

    return list(set(results))


def all_possible_show_names(series_id, series_provider_id, season=-1):
    """
    Figures out every possible variation of the name for a particular show. Includes TVDB name, TVRage name,
    country codes on the end, eg. "Show Name (AU)", and any scene exception names.

    show: a TVShow object that we should get the names of

    Returns: a list of all the possible show names
    :rtype: list[unicode]
    """

    show = find_show(series_id, series_provider_id)

    show_names = show.get_scene_exceptions_by_season(season=season)[:]
    if not show_names:  # if we dont have any season specific exceptions fallback to generic exceptions
        season = -1
        show_names = show.get_scene_exceptions_by_season(season=season)[:]

    if season in [-1, 1]:
        show_names.append(show.name)

    show_names.append(strip_accents(show.name))
    show_names.append(strip_accents(show.name).replace("'", " "))

    if not show.is_anime:
        new_show_names = []
        country_list = countryList
        country_list.update(dict(zip(countryList.values(), countryList.keys())))
        for curName in set(show_names):
            if not curName:
                continue

            # if we have "Show Name Australia" or "Show Name (Australia)" this will add "Show Name (AU)" for
            # any countries defined in common.countryList
            # (and vice versa)
            for curCountry in country_list:
                if curName.endswith(' ' + curCountry):
                    new_show_names.append(curName.replace(' ' + curCountry, ' (' + country_list[curCountry] + ')'))
                elif curName.endswith(' (' + curCountry + ')'):
                    new_show_names.append(curName.replace(' (' + curCountry + ')', ' (' + country_list[curCountry] + ')'))

            # if we have "Show Name (2013)" this will strip the (2013) show year from the show name
            new_show_names.append(re.sub(r'\({}\)'.format(show.startyear), '', curName))

        show_names += new_show_names

    return list(set(show_names))


def determine_release_name(dir_name=None, nzb_name=None):
    """Determine a release name from an nzb and/or folder name
    :param dir_name:
    :param nzb_name:
    :return:
    """

    if nzb_name is not None:
        sickrage.app.log.info("Using nzb_name for release name.")
        return nzb_name.rpartition('.')[0]

    if dir_name is None:
        return None

    # try to get the release name from nzb/nfo
    file_types = ["*.nzb", "*.nfo"]

    for search in file_types:
        reg_expr = re.compile(fnmatch.translate(search), re.IGNORECASE)
        files = [file_name for file_name in os.listdir(dir_name) if
                 os.path.isfile(os.path.join(dir_name, file_name))]

        results = list(filter(reg_expr.search, files))
        if len(results) == 1:
            found_file = os.path.basename(results[0])
            found_file = found_file.rpartition('.')[0]
            if filter_bad_releases(found_file):
                sickrage.app.log.info("Release name ({}) found from file ({})".format(found_file, results[0]))
                return found_file.rpartition('.')[0]

    # If that fails, we try the folder
    folder = os.path.basename(dir_name)
    if filter_bad_releases(folder):
        # NOTE: Multiple failed downloads will change the folder name.
        # (e.g., appending #s)
        # Should we handle that?
        sickrage.app.log.debug("Folder name (" + folder + ") appears to be a valid release name. Using it.")
        return folder

    return None



"""Initial migration

Revision ID: 3
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '3'
down_revision = '2'


def upgrade():
    with op.batch_alter_table('search_providers_newznab') as batch_op:
        batch_op.drop_column('key')


def downgrade():
    pass



# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
from xml.etree.ElementTree import Element, ElementTree, SubElement

import sickrage
from sickrage.core.common import dateFormat
from sickrage.core.helpers import indent_xml
from sickrage.metadata_providers import MetadataProvider
from sickrage.series_providers.exceptions import SeriesProviderEpisodeNotFound, \
    SeriesProviderSeasonNotFound


class KODI_12PlusMetadata(MetadataProvider):
    """
    Metadata generation class for KODI 12+.

    The following file structure is used:

    show_root/tvshow.nfo                    (show metadata)
    show_root/fanart.jpg                    (fanart)
    show_root/poster.jpg                    (poster)
    show_root/banner.jpg                    (banner)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.nfo        (episode metadata)
    show_root/Season ##/filename-thumb.jpg  (episode thumb)
    show_root/season##-poster.jpg           (season posters)
    show_root/season##-banner.jpg           (season banners)
    show_root/season-all-poster.jpg         (season all poster)
    show_root/season-all-banner.jpg         (season all banner)
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False):

        MetadataProvider.__init__(self,
                                  show_metadata,
                                  episode_metadata,
                                  fanart,
                                  poster,
                                  banner,
                                  episode_thumbnails,
                                  season_posters,
                                  season_banners,
                                  season_all_poster,
                                  season_all_banner)

        self.name = 'KODI 12+'

        self.poster_name = "poster.jpg"
        self.season_all_poster_name = "season-all-poster.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "tvshow.nfo"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.nfo"
        self.eg_fanart = "fanart.jpg"
        self.eg_poster = "poster.jpg"
        self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>-thumb.jpg"
        self.eg_season_posters = "season##-poster.jpg"
        self.eg_season_banners = "season##-banner.jpg"
        self.eg_season_all_poster = "season-all-poster.jpg"
        self.eg_season_all_banner = "season-all-banner.jpg"

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for an KODI-style tvshow.nfo and
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        tv_node = Element("tvshow")

        series_provider_language = show_obj.lang or sickrage.app.config.general.series_provider_default_language
        series_info = show_obj.series_provider.get_series_info(show_obj.series_id, language=series_provider_language)
        if not series_info:
            return False

        # check for title and id
        if not (getattr(series_info, 'name', None) and getattr(series_info, 'id', None)):
            sickrage.app.log.info("Incomplete info for show with id " + str(show_obj.series_id) + " on " + show_obj.series_provider.name + ", skipping it")
            return False

        title = SubElement(tv_node, "title")
        title.text = series_info["name"]

        if getattr(series_info, 'rating', None):
            rating = SubElement(tv_node, "rating")
            rating.text = str(series_info["rating"])

        if getattr(series_info, 'firstAired', None):
            try:
                year_text = str(datetime.datetime.strptime(series_info["firstAired"], dateFormat).year)
                if year_text:
                    year = SubElement(tv_node, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(series_info, 'overview', None):
            plot = SubElement(tv_node, "plot")
            plot.text = series_info["overview"]

        # if getattr(series_info, 'id', None):
        #    episodeguide = SubElement(tv_node, "episodeguide")
        #    episodeguideurl = SubElement(episodeguide, "url")
        #    episodeguideurl.text = IndexerApi(show_obj.series_provider_id).config['base_url'] + str(
        #        series_info["id"]) + '/all/en.zip'

        # if getattr(series_info, 'contentrating', None):
        #     mpaa = SubElement(tv_node, "mpaa")
        #     mpaa.text = series_info["contentrating"]

        if getattr(series_info, 'id', None):
            series_id = SubElement(tv_node, "id")
            series_id.text = str(series_info["id"])

        if getattr(series_info, 'genres', None):
            genre = SubElement(tv_node, "genre")
            genre.text = " / ".join(x['name'] for x in series_info["genres"])

        if getattr(series_info, 'firstAired', None):
            premiered = SubElement(tv_node, "premiered")
            premiered.text = series_info["firstAired"]

        if getattr(series_info, 'network', None):
            studio = SubElement(tv_node, "studio")
            studio.text = series_info["network"].strip()

        for person in series_info['people']:
            if 'name' not in person or not person['name'].strip():
                continue

            if person['role'].strip() == 'Actor':
                cur_actor = SubElement(tv_node, "actor")

                cur_actor_role = SubElement(cur_actor, "role")
                cur_actor_role.text = person['role'].strip()

                cur_actor_name = SubElement(cur_actor, "name")
                cur_actor_name.text = person['name'].strip()

                if person['imageUrl'].strip():
                    cur_actor_thumb = SubElement(cur_actor, "thumb")
                    cur_actor_thumb.text = person['imageUrl'].strip()

        # Make it purdy
        indent_xml(tv_node)

        data = ElementTree(tv_node)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for an KODI-style episode.nfo and
        returns the resulting data object.
            show_obj: a TVEpisode instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.related_episodes

        series_provider_language = ep_obj.show.lang or sickrage.app.config.general.series_provider_default_language
        series_info = ep_obj.show.series_provider.get_series_info(ep_obj.show.series_id, language=series_provider_language)
        if not series_info:
            return False

        if len(eps_to_write) > 1:
            root_node = Element("kodimultiepisode")
        else:
            root_node = Element("episodedetails")

        # write an NFO containing info for all matching episodes
        for curEpToWrite in eps_to_write:
            try:
                series_episode_info = series_info[curEpToWrite.season][curEpToWrite.episode]
            except (SeriesProviderEpisodeNotFound, SeriesProviderSeasonNotFound):
                sickrage.app.log.info(
                    f"Unable to find episode {curEpToWrite.season:d}x{curEpToWrite.episode:d} on {ep_obj.show.series_provider.name}"
                    f"... has it been removed? Should I delete from db?")
                return None

            if not getattr(series_episode_info, 'firstAired', None):
                series_episode_info["firstAired"] = str(datetime.date.min)

            if not getattr(series_episode_info, 'name', None):
                sickrage.app.log.debug("Not generating nfo because the ep has no title")
                return None

            sickrage.app.log.debug(f"Creating metadata for episode {ep_obj.season}x{ep_obj.episode}")

            if len(eps_to_write) > 1:
                episode = SubElement(root_node, "episodedetails")
            else:
                episode = root_node

            if getattr(series_episode_info, 'name', None):
                title = SubElement(episode, "title")
                title.text = series_episode_info['name']

            if getattr(series_info, 'name', None):
                showtitle = SubElement(episode, "showtitle")
                showtitle.text = series_info['name']

            season = SubElement(episode, "season")
            season.text = str(curEpToWrite.season)

            episodenum = SubElement(episode, "episode")
            episodenum.text = str(curEpToWrite.episode)

            uniqueid = SubElement(episode, "uniqueid")
            uniqueid.text = str(curEpToWrite.series_id)

            if curEpToWrite.airdate > datetime.date.min:
                aired = SubElement(episode, "aired")
                aired.text = str(curEpToWrite.airdate)

            if getattr(series_episode_info, 'overview', None):
                plot = SubElement(episode, "plot")
                plot.text = series_episode_info['overview']

            if curEpToWrite.season and getattr(series_info, 'runtime', None):
                runtime = SubElement(episode, "runtime")
                runtime.text = series_info["runtime"]

            if getattr(series_episode_info, 'airsbefore_season', None):
                displayseason = SubElement(episode, "displayseason")
                displayseason.text = series_episode_info['airsbefore_season']

            if getattr(series_episode_info, 'airsbefore_episode', None):
                displayepisode = SubElement(episode, "displayepisode")
                displayepisode.text = series_episode_info['airsbefore_episode']

            if getattr(series_episode_info, 'filename', None):
                thumb = SubElement(episode, "thumb")
                thumb.text = series_episode_info['filename'].strip()

            # watched = SubElement(episode, "watched")
            # watched.text = 'false'

            if getattr(series_episode_info, 'rating', None):
                rating = SubElement(episode, "rating")
                rating.text = str(series_episode_info['rating'])

            for person in series_info['people']:
                if 'name' not in person or not person['name'].strip():
                    continue

                if person['role'].strip() == 'Actor':
                    cur_actor = SubElement(episode, "actor")

                    cur_actor_role = SubElement(cur_actor, "role")
                    cur_actor_role.text = person['role'].strip()

                    cur_actor_name = SubElement(cur_actor, "name")
                    cur_actor_name.text = person['name'].strip()

                    if person['imageUrl'].strip():
                        cur_actor_thumb = SubElement(cur_actor, "thumb")
                        cur_actor_thumb.text = person['imageUrl'].strip()
                elif person['role'].strip() == 'Writer':
                    ep_credits = SubElement(episode, "credits")
                    ep_credits.text = series_episode_info['writer'].strip()
                elif person['role'].strip() == 'Director':
                    director = SubElement(episode, "director")
                    director.text = series_episode_info['director'].strip()
                elif person['role'].strip() == 'Guest Star':
                    cur_actor = SubElement(episode, "actor")
                    cur_actor_name = SubElement(cur_actor, "name")
                    cur_actor_name.text = person['name'].strip()

        # Make it purdy
        indent_xml(root_node)
        data = ElementTree(root_node)
        return data



# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


from urllib.parse import urlencode

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.enums import SearchFormat
from sickrage.core.exceptions import AuthException
from sickrage.core.helpers import try_int
from sickrage.core.tv.show.helpers import find_show
from sickrage.search_providers import TorrentProvider


class HDBitsProvider(TorrentProvider):
    def __init__(self):
        super(HDBitsProvider, self).__init__("HDBits", 'https://hdbits.org', True)

        # custom settings
        self.custom_settings = {
            'username': '',
            'passkey': '',
        }

        self.cache = HDBitsCache(self, min_time=15)

    @property
    def urls(self):
        return {
            'search': f'{self.url}/api/torrents',
            'rss': f'{self.url}/api/torrents',
            'download': f'{self.url}/download.php'
        }

    def _check_auth(self):
        if not self.custom_settings['username'] or not self.custom_settings['passkey']:
            raise AuthException("Your authentication credentials for " + self.name + " are missing, check your config.")

        return True

    def _check_auth_from_data(self, parsed_json):
        if 'status' in parsed_json and 'message' in parsed_json:
            if parsed_json.get('status') == 5:
                sickrage.app.log.warning(
                    "Invalid username or password. Check your settings")

        return True

    def _get_season_search_strings(self, series_id, series_provider_id, season, episode):
        post_data = {
            'username': self.custom_settings['username'],
            'passkey': self.custom_settings['passkey'],
            'category': [2],
            # TV Category
        }

        show_object = find_show(series_id, series_provider_id)
        if not show_object:
            return [post_data]

        episode_object = show_object.get_episode(season, episode)

        if show_object.search_format in [SearchFormat.AIR_BY_DATE, SearchFormat.SPORTS]:
            post_data['tvdb'] = {
                'id': series_id,
                'season': str(episode_object.airdate)[:7],
            }
        elif show_object.search_format == SearchFormat.ANIME:
            post_data['tvdb'] = {
                'id': series_id,
                'season': "%d" % episode_object.get_absolute_numbering(),
            }
        else:
            post_data['tvdb'] = {
                'id': series_id,
                'season': episode_object.get_season_episode_numbering()[0],
            }

        return [post_data]

    def _get_episode_search_strings(self, series_id, series_provider_id, season, episode, add_string=''):
        post_data = {
            'username': self.custom_settings['username'],
            'passkey': self.custom_settings['passkey'],
            'category': [2],
            # TV Category
        }

        show_object = find_show(series_id, series_provider_id)
        if not show_object:
            return [post_data]

        episode_object = show_object.get_episode(season, episode)

        if show_object.search_format == SearchFormat.AIR_BY_DATE:
            post_data['tvdb'] = {
                'id': series_id,
                'episode': str(episode_object.airdate).replace('-', '|')
            }
        elif show_object.search_format == SearchFormat.SPORTS:
            post_data['tvdb'] = {
                'id': series_id,
                'episode': episode_object.airdate.strftime('%b')
            }
        elif show_object.search_format == SearchFormat.ANIME:
            post_data['tvdb'] = {
                'id': series_id,
                'episode': "%i" % episode_object.get_absolute_numbering()
            }
        else:
            post_data['tvdb'] = {
                'id': series_id,
                'season': episode_object.get_season_episode_numbering()[0],
                'episode': episode_object.get_season_episode_numbering()[1]
            }

        return [post_data]

    def _get_title_and_url(self, item):
        title = item['name']
        if title:
            title = self._clean_title_from_provider(title)

        url = self.urls['download'] + '?' + urlencode({'id': item['id'], 'passkey': self.custom_settings['passkey']})

        return title, url

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        sickrage.app.log.debug("Search string: %s" % search_strings)

        self._check_auth()

        resp = self.session.post(self.urls['search'], json=search_strings)
        if not resp or resp.content:
            sickrage.app.log.warning("Resulting JSON from provider isn't correct, not parsing it")
            return results

        try:
            parsed_json = resp.json()
        except ValueError:
            sickrage.app.log.warning("Resulting JSON from provider isn't correct, not parsing it")
            return results

        if self._check_auth_from_data(parsed_json):
            if not parsed_json or 'data' not in parsed_json:
                sickrage.app.log.warning("Resulting JSON from provider isn't correct, not parsing it")
                return results

            for item in parsed_json['data']:
                results.append(item)

        # sort by number of seeders
        results.sort(key=lambda k: try_int(k.get('seeders', 0)), reverse=True)

        return results


class HDBitsCache(TVCache):
    def _get_rss_data(self):
        results = []

        post_data = {
            'username': self.provider.custom_settings['username'],
            'passkey': self.provider.custom_settings['passkey'],
            'category': [2],
        }

        resp = self.provider.session.post(self.provider.urls['rss'], json=post_data)
        if not resp or not resp.content:
            return results

        try:
            parsed_json = resp.json()
        except ValueError:
            return results

        if self.provider._check_auth_from_data(parsed_json):
            results = parsed_json['data']

        return {'entries': results}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: updates_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10updates_v1.proto\x12\x10\x61pp.protobufs.v1\"#\n\x12UpdatedAppResponse\x12\r\n\x05\x66orce\x18\x01 \x01(\x08\x62\x06proto3')



_UPDATEDAPPRESPONSE = DESCRIPTOR.message_types_by_name['UpdatedAppResponse']
UpdatedAppResponse = _reflection.GeneratedProtocolMessageType('UpdatedAppResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEDAPPRESPONSE,
  '__module__' : 'updates_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.UpdatedAppResponse)
  })
_sym_db.RegisterMessage(UpdatedAppResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _UPDATEDAPPRESPONSE._serialized_start=38
  _UPDATEDAPPRESPONSE._serialized_end=73
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: server_certificate_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bserver_certificate_v1.proto\x12\x10\x61pp.protobufs.v1\"J\n\x1eSavedServerCertificateResponse\x12\x13\n\x0b\x63\x65rtificate\x18\x01 \x01(\t\x12\x13\n\x0bprivate_key\x18\x02 \x01(\tb\x06proto3')



_SAVEDSERVERCERTIFICATERESPONSE = DESCRIPTOR.message_types_by_name['SavedServerCertificateResponse']
SavedServerCertificateResponse = _reflection.GeneratedProtocolMessageType('SavedServerCertificateResponse', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDSERVERCERTIFICATERESPONSE,
  '__module__' : 'server_certificate_v1_pb2'
  # @@protoc_insertion_point(class_scope:app.protobufs.v1.SavedServerCertificateResponse)
  })
_sym_db.RegisterMessage(SavedServerCertificateResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SAVEDSERVERCERTIFICATERESPONSE._serialized_start=49
  _SAVEDSERVERCERTIFICATERESPONSE._serialized_end=123
# @@protoc_insertion_point(module_scope)
